/* Project:     HARTIK 3.0 Network Library                      */
/* Description: Hard Real TIme Kernel for 8086 compatible       */
/* Author:      Luca Abeni					*/
/* Date:        25/9/1997                                       */

/* File: 	NetBuff.c					*/
/* Revision:    1.00						*/

/*
   This module manages all the buffers used by the network driver:
   the buffers are grouped in netbuff structures (initialized by
   netbuff_init) and can be allocated by netbuff_get or released
   by netbuff_released. A netbuff_sequentialget function is also
   provided for beeing used by tasks that have their own buffer (
   if a buffer can be allocated by only one task, it is useless to
   allocate it in mutual exclusion). It is usefull for the receive
   buffers allocated by the receive process.
*/
#include <hartik.h>
#include <string.h>
#include <vm.h>

#include "netbuff.h"

/* Init the buffer structures */
void netbuff_init(struct netbuff *netb, BYTE nbuffs, WORD buffdim)
{
    int i;

    VM_disable();
    netb->b = VM_alloc(nbuffs * buffdim);
    netb->pb = VM_alloc(nbuffs * sizeof(void *));
    netb->free = VM_alloc(nbuffs * sizeof(char));
    VM_enable();
    if ((netb->b == 0) || (netb->pb ==0) || (netb->free == 0)) {
	exc_raise(NETBUFF_INIT_EXC);
    }
    netb->nbuffs = nbuffs;
    for (i = 0; i < nbuffs; i++) {
	netb->pb[i] = netb->b + (i * buffdim);
	netb->free[i] = 1;
    }
    netb->buffersem = sem_create(nbuffs);
}

/*
   Get the first free buffer in the netb pool. All the tasks can call this
   function, so a VM_disable is needed
*/
void *netbuff_get(struct netbuff *netb, BYTE to)
{
    int i, done;

    if (!sem_wait(netb->buffersem, to)) {
	return NULL;
    }

    done = 0; i = 0;
    VM_disable();
    while ((i < netb->nbuffs) && !done) {
	if (netb->free[i]) {
	    done = 1;
	    netb->free[i] = 0;
	} else i++;
    }
    VM_enable();
    if (!done) {
	exc_raise(NETBUFF_GET_EXC);
    }
    return netb->pb[i];
}

/*
   Get the first free buffer in the netb pool. This function can be called
   only if netb is private of the process that wants to get the buffer, so
   VM_disable is not needed
*/
void *netbuff_sequentialget(struct netbuff *netb, BYTE to)
{
    int i, done;

    if (!sem_wait(netb->buffersem, to)) {
	return NULL;
    }

    done = 0; i = 0;
    while ((i < netb->nbuffs) && !done) {
	if (netb->free[i]) {
	    done = 1;
	    netb->free[i] = 0;
	} else i++;
    }
    if (!done) {
	exc_raise(NETBUFF_GET_EXC);
    }
    return netb->pb[i];
}

/* Release the buffer m of the pool netb */
void netbuff_release(struct netbuff *netb, void *m)
{
    int i, done;

    done = 0; i = 0;
    while ((i < netb->nbuffs) && !done) {
	if (netb->pb[i] == m) {
	    if (netb->free[i] == 1) {
		exc_raise(NETBUFF_ALREADYFREE_EXC);
	    }
	    done = 1;
	    netb->free[i] = 1;
	} else i++;
    }
    if (!done) {
	exc_raise(NETBUFF_RELEASE_EXC);
    }
    sem_signal(netb->buffersem);
}
