/* Project:     HARTIK 3.0 Sound Library                        */
/* Description: Hard Real TIme Kernel for 8086 compatible       */
/* Author:      Luca Abeni					*/
/* Date:        5/12/1997                                       */

/* File: 	Sound.C						*/
/* Revision:    3.0						*/

/* Sound generic interface */
#include <hartik.h>
#include <kern.h>
#include <vm.h>

#include "cons.h"

#include "blaster.h"
#include "sound.h"
#include "dma.h"

extern struct proc_des proc_table[MAX_PROC];

extern struct sb_device sb_dev;
extern TASK proc_play();
extern TASK proc_sample();
extern TASK sb_proc();
extern void sb_handler(int n);
extern struct sound_buffer buff_des;
struct rawfuncs f_des;
struct dma_buff dma_des;
WORD sound_tick;

int sound_exc(int err)
{
    switch (err) {
	case SOUND_CARD_NOT_FOUND:
	    cprintf("SOUND PANIC --> Sound card not found.\n");
	    return 0;
	case SOUND_RAWFUN_NOT_SET:
	    cprintf("Self-buffering OP error --> User function not set.\n");
	    return 0;
	case SOUND_TICK_TOO_LARGE:
	    cprintf("PIO mode OP error --> sys_tick too large.\n");
	    return 0;
	case SOUND_ASYNCH_WAIT:
	    cprintf("Asynchronous OP error --> Wait on asynch operation.\n");
	    return 1;
	default :
	    return 1;
    }
}

/* Set the user's functions for self buffering */
void sound_setfun(int (*infun)(BYTE *rawbuff), int (*outfun)(BYTE *rawbuff))
{
    /* Should be NULL here! Luca: used -1 instead NULL */
    if (infun != NULL) {
	f_des.infun = infun;
	f_des.infunpresent = 1;
    }
    if(outfun != NULL) {
	f_des.outfun = outfun;
	f_des.outfunpresent = 1;
    }
}

/*
   This function is called by the sound card driver function when the card
   generate an interrupt during a self buffering input operation. The user's
   self buffering function is called by this function
*/
int selfinfun(struct dma_buff *b)
{
    int res;
    
    res = f_des.infun(&b->dma_buff[(b->dma_bufflen >> 1) * b->page]);
    b->page = !b->page;
    return res;
}

/*
   This function is called by the sound card driver function when the card
   generate an interrupt during a self buffering output operation. The user's
   self buffering function is called by this function
*/
int selfoutfun(struct dma_buff *b)
{
    int res;
    
    res = f_des.outfun(&b->dma_buff[(b->dma_bufflen >> 1) * b->page]);
    b->page = !b->page;
    return res;
}

/*
   Play a sample: this function calls the adeguate low-level playing function
   with the correct parameters
*/
void sound_play(BYTE *buff, DWORD sps, DWORD len, BYTE t)
{
    double rate;
    PID p_sb,p_pl;
    WORD s1, p;
    MODEL m = BASE_MODEL;

    /* Self Buffering?*/
    if (t & MYFUN) {
	/*Yes */
	if (f_des.outfunpresent != 0) {
	    buff_des.fun = selfoutfun;
	} else exc_raise(SOUND_RAWFUN_NOT_SET);
    }
    else {
	/* No: set the standard double-buffering output function (module dma.c) */
	buff_des.fun = outfun;
    }
    /* set the operation parameters...*/
    buff_des.synch = !(t & SYNCH);
    dma_des.len = len;
    dma_des.p = buff;
    if (t & DMA_OP) {
	/* DMA op */
	sb_setrate(sps, OUT);
	if (t & DYNBUFF) {
	    /* Compute the driver task's minimum interarrival time */
	    p_sb = task_pid("sb_EndDMA");
	    s1 = (sps / 1000) * ((t & PCM16) ? 2 : 1);
	    p = ((dma_des.dma_bufflen >> 1) * 1000 ) / (sound_tick * s1);
	    if(p < 1) {
		sb_dev.pwarning = 1;
		p = 1;
	    } else sb_dev.pwarning = 0;
	    /* Adjust the driver task's minimum interarrival time */
	    VM_disable();
	    proc_table[p_sb].period = p;
	    proc_table[p_sb].drel = p;
	    VM_enable();
	    sb_dev.period = p;
	}
	if (!(t & NOBUFF)) {
	    if (t & PCM16) sb_dma16buffop(OUT);
	    else sb_dmabuffop(OUT);
	} else {
	    /* Double buffering operation */
	    buff_des.fun = dummyfun2;
	    if (t & PCM16) sb_dma16op(OUT);
	    else sb_dmaop(OUT);
	}
    } else {
	/* Non DMA op: create the playing process */
	rate = ((1000000 / sound_tick) + (sps / 2)) / sps;
	if (rate < 1) exc_raise(SOUND_TICK_TOO_LARGE);
	task_def_wcet(m,10);
	p_pl = task_create("sb_Player",proc_play, HARD, PERIODIC, rate, &m);
	if (p_pl == NIL) {
	    VM_printf("Sound.c: Cannot create sb_Player\n");
	    sys_end();
	    exit(-1);
	}
	task_activate(p_pl);
    }
}

/*
   Sample: this function calls the adeguate low-level sampling function
   with the correct parameters. It is similar to sound_play (see it for the
   comments
*/
void sound_sample(BYTE *buff, DWORD sps, DWORD len, BYTE t)
{
    double rate;
    PID p_sb,p_sample;
    WORD s1, p;
    MODEL m = BASE_MODEL;

    if (t & MYFUN) 
	if (f_des.infunpresent != 0) {
	    buff_des.fun = selfinfun;
	} else exc_raise(SOUND_RAWFUN_NOT_SET);
    else {
	buff_des.fun = infun;
    }
    buff_des.synch = !(t & SYNCH);
    dma_des.len = len;
    dma_des.p = buff;
    if (t & DMA_OP) {
	sb_setrate(sps, IN);
	if (t & DYNBUFF) {
	    p_sb = task_pid("sb_EndDMA");
	    s1 = (sps / 1000) * ((t & PCM16) ? 2 : 1);
	    p = ((dma_des.dma_bufflen >> 1) * 1000 ) / (sound_tick * s1);
	    if(p < 1) {
		sb_dev.pwarning = 1;
		p = 1;
	    } else sb_dev.pwarning = 0;
	    VM_disable();
	    proc_table[p_sb].period = p;
	    proc_table[p_sb].drel = p;
	    VM_enable();
	    sb_dev.period = p;
	}
	if (!(t & NOBUFF)) {
	    if (t & PCM16) sb_dma16buffop(IN);
	    else sb_dmabuffop(IN);
	} else {
	    buff_des.fun = dummyfun2;
	    if (t & PCM16) sb_dma16op(IN);
	    else sb_dmaop(IN);
	}
    } else {
	rate = ((1000000 / sound_tick) + (sps / 2)) / sps;
	if (rate < 1) exc_raise(SOUND_TICK_TOO_LARGE);
	task_def_wcet(m,10);
	p_sample = task_create("sb_Sampler",proc_sample, HARD, PERIODIC, rate, &m);
	if (p_sample == NIL) {
	    VM_printf("Sound.c: Cannot create sb_Sampler\n");
	    sys_end();
	    exit(-1);
	}
	task_activate(p_sample);
    }
}

/*
   Wait for the end of a synchronous sound op: it is implemented with a
   simple semaphore
*/
void sound_wait(void)
{
    if (buff_des.synch) {
	sem_wait(buff_des.synchr,BLOCK);
	buff_des.synch = 0;
    } else exc_raise(SOUND_ASYNCH_WAIT);
}

void sound_init(WORD rawbuffsize, WORD tick)
{
    MODEL m = BASE_MODEL;
    PID p_sb;
    int i,period;

    for (i = SOUND_ERROR_BASE; i <= SOUND_ASYNCH_WAIT; i++)
	exc_set(i,sound_exc);

    sound_tick = tick;
    /* Semaphore for synchronous ops */
    buff_des.synchr = sem_create(0);
    /* Init the card */
    sb_init();
    sbmixer_reset();
    sbmixer_setinput(0x01, ENABLE);
    sbmixer_setoutput(0x01, DISABLE);
    sbmixer_setmiclev(0x1F);
    sbmixer_setAGC(ENABLE);
    sbmixer_setingainlev(0);

    f_des.infun = 0;
    f_des.infunpresent = 0;
    f_des.outfun = 0;
    f_des.outfunpresent = 0;
    /* init the buffers for DMA ops */
    dma_getalignbuff(&dma_des, rawbuffsize);
    buff_des.sound_dma = &dma_des;
    period = ((rawbuffsize >> 1) * 12) / sound_tick;
    if(period < 1) {
	sb_dev.pwarning = 1;
	period = 1;
    } else sb_dev.pwarning = 0;
    /* create the driver process and set it and the Fast Handler */
    task_def_wcet(m,10);
    p_sb = task_create("sb_EndDMA", sb_proc, SOFT, APERIODIC, period, &m);
    if (p_sb == NIL) {
	VM_printf("Sound.c: Cannot create sb_EndDMA\n");
	VM_printf("Period = %u Wcet = %u\nError : %s\n",period,m.wcet,sys_err(errno()));
	sys_end();
	exit(-1);
    }
    sb_dev.period = period;
    handler_set(sb_dev.IntLine, sb_handler, p_sb);
}

/* Obvious... */
void sound_info(void)
{
    cprintf("\n\n                Hartik Sound lib [V 3.2]:\n\n");
    cprintf("Sound Blaster 16 or clone found:\n");
    sb_show();
    cprintf("Initial server period: %d\n",sb_dev.period);
    if (sb_dev.pwarning) cprintf("Warning: sys_tick too big!!!\n");
}

/* Return the minimum intererrival time of the driver process */
WORD sound_serverp(void)
{
    return sb_dev.period;
}

/*
   This function is called by the driver process on the last transfert of an
   operation
*/
int dummyfun1(struct dma_buff *d)
{
    buff_des.fun = dummyfun2;
    return 0;
}

/*
   This function is called by the driver process when the next transfert
   will be the last of the current operation
*/
int dummyfun2(struct dma_buff *d)
{
    if (buff_des.synch) sem_signal(buff_des.synchr);
    sb_stopdsp(8);
    sb_stopdsp(16);
    dma_stop(sb_dev.DMA8Channel);
    dma16_stop(sb_dev.DMA16Channel);
    return 0;
}

/* Obvious... */
void sound_stop(void)
{
    if (!buff_des.synch) {
	sb_stopdsp(8);
	sb_stopdsp(16);
	dma_stop(sb_dev.DMA8Channel);
	dma16_stop(sb_dev.DMA16Channel);
    }
}
