/* Project:     HARTIK 3.0                                      */
/* Description: Hard Real TIme Kernel for 8086 compatible       */
/* Author:      Gerardo Lamastra                                */
/*              Giuseppe Lipari                                 */
/* Date:        1/7/96                                          */

/* CVS $Id: kern.h,v 1.5 1999/11/04 10:15:19 massy0 Exp $ */

/* File:        Kern.H                                          */
/* Revision:    3.0                                             */
/* Date:	14/3/97						*/

/* Kernel main data structures & primitives			*/

#ifndef __KERN_H__
#define __KERN_H__

/* Autoinclude all the necessary stuff! */

#ifndef __HW_DEP_H__
#include "hw_dep.h"
#endif

#ifndef __CONFIG_H__
#include "config.h"
#endif

#ifndef __CONST_H__
#include "const.h"
#endif

#ifndef __EXC_H__
#include "exc.h"
#endif

#ifndef __MODEL_H__
#include "model.h"
#endif

#ifndef __SYS_H__
#include "sys.h"
#endif

/* Some useful macros */
#define FOREVER         for(;;)
#define EMPTY(q)        ((q) == NIL)
#define PRIO2DLINE(x)   ((DWORD)(MAX_DLINE-(MAX_PRIORITY-x)))

/* The CBS standard prologue: this macro is called by each syscall that
   can perform a context switch...					*/
#define CBS_PROLOGUE() {						\
    act_time = VM_timer_read();						\
									\
    proc_table[exec].cap -= (act_time - last_int);			\
    last_int = act_time;		/* update the last int time */	\
}									\



/*---------------------------------------------------------------------*/
/*                       TASK DESCRIPTOR                               */
/*---------------------------------------------------------------------*/
/* Field explanation:                                                  */
/* <context> is the pointer to the context of the task (see vm.h)      */
/* <body> is the starting address of the task                          */
/*                                                                     */
/* <status> represent the actual task status:			   */
/*			 (EXE,READY,IDLE,FREE,SLEEP)			   */
/*                                                                     */
/* <type,pclass> identify the task model                               */
/*                                                                     */
/* <group> is 0 if the task does not belong to a group; elsewhere it   */
/* identifies the task group id                                        */
/*                                                                     */
/* <dline,period> are used to describe the task carachteristichs       */
/*                                                                     */
/* <next,prev> are used to insert the tasks into queues                */

#define NO_PREEMPT	0x8000
struct proc_des {
	CONTEXT context;   /* Context area pointer         	    */
	BYTE *stack;       /* Pointer to stack area base            */
	TASK (*body)();    /* Pointer to the code of the task       */
	char name[20];     /* Text identifing the process name      */
	WORD status;       /* This could be READY,EXE,FREE ...      */
	WORD type;         /* PERIODIC,APERIODIC		    */
	WORD pclass;       /* HARD,FIRM,(NEW)SOFT,NRT               */
	WORD group;        /* 0 if task is single, else group id    */
	WORD stacksize;    /* Task stack size                       */
	WORD control;      /* Control task operating mode           */
			   /* Refer to the MODEL type for its use   */
	DWORD priority;    /* Used to sort the task queues	    */
	DWORD dline;       /* Absolute deadline of any task;	    */
			   /* for NRT we translate priority	    */
			   /* into deadline 			    */
	DWORD drel;        /* Relative deadline                     */
	DWORD period;      /* Period of periodic tasks              */
			   /* Priority of NRT task                  */
	DWORD wcet;        /* Worst Case Exceution Time             */
	DWORD num_act;	   /* Number of pending activations         */
	DWORD c_act;       /* Actual exceution time                 */
	DWORD c_resid;     /* Residual worst case execution time    */
	DWORD request;     /* Last request time for the task        */
	QUEUE next,prev;   /* Next/Prev Index in the queue          */

	/*these fields was been added by Luca for CBS support...    */
	long int cap;	   /* CBS max capacity			    */
	DWORD q;	   /* CBS actual capacity		    */
	DWORD last;	   /* last CBS capacity management time	    */
	DWORD *exect;

	BYTE errno;
};

/*---------------------------------------------------------------------*/
/* 		Load Struct (by Luca...)			       */
/*---------------------------------------------------------------------*/

struct load {
    DWORD actual;
    DWORD rif;
    DWORD newrif;
};


/*---------------------------------------------------------------------*/
/* Kernel global variables: system queues & clock counters ...         */
/*---------------------------------------------------------------------*/

extern struct proc_des proc_table[];
extern PID exec;
extern QUEUE freedesc;		/* Free descriptor handled as a queue	*/
extern QUEUE ready_rt;		/* Ready queue for real-time tasks	*/
extern QUEUE ready_nrt; 	/* Ready queue for the others		*/
extern QUEUE idle;		/* Idle periodic task queues		*/
extern QUEUE blocked;		/* Blocked queue                	*/
extern QUEUE free_sem;		/* Queue of free sem            	*/

extern DWORD sys_clock;		/* System clock (in tick numbers)	*/
extern TIME  sys_tick;		/* System tick (in usec)		*/

/* some information macros */
#define task_getperiod(t) (proc_table[(t)].period)
#define task_getdrel(t)   (proc_table[(t)].drel)
#define task_getdline(t)  (proc_table[(t)].dline)

#define task_setperiod(t,p) {\
   if ((t)>=0 && (t)<MAX_PROC) \
      proc_table[(t)].period=(p); \
}
#define task_setdrel(t,p) {\
   if ((t)>=0 && (t)<MAX_PROC) \
      proc_table[(t)].drel=(p); \
}
#define task_setdline(t,p) {\
   if ((t)>=0 && (t)<MAX_PROC) \
      proc_table[(t)].dline=(p); \
}

/* Indicates if there is an active fast routine for interrupt 	*/
/* It is necessary for the scheduler to know if an interrupt  	*/
/* is active							*/
extern BYTE isActiveInt;

/* The TBS allocated bandwidth! Another system parameter! */
/* The bandwidth is an integral; the effective value is	  */
/* TBS_util / TBS_PREC!					  */
/*
   NOW OBSOLETE!!!
   #define TBS_PREC 	1000
   extern DWORD TBS_util;
*/
#define MIN_Q 20

/* Process descriptor table */
extern struct proc_des proc_table[];

/* Queue management functions */
void q_insert(QUEUE p, QUEUE *q);
QUEUE q_extract(QUEUE p, QUEUE *q);
QUEUE q_getfirst(QUEUE *q);

/* This function implements the EDF scheduling strategy  */
void scheduler(void);
void dispatch(void);

/* Timer interrupt service routine */
CONTEXT __cdecl wake_up(CONTEXT c);

/* Generic interrupt service routine */
CONTEXT __cdecl act_int(CONTEXT c,BYTE n);

/* Generic exception trapping routine */
void __cdecl act_exc(int code);

/* Init semaphoric struct */
void sem_init(void);

/* Initialize both the interrupt and the exception table */
void excirq_init(void);

/* Initialize the load monitor */
void load_init(void);

#endif /* __KERN_H__ */
