/* Project:     HARTIK 3.0                                      */
/* Description: Hard Real TIme Kernel for 386 & higher machines */
/* Author:      Gerardo Lamastra                                */
/*              Paolo della Capanna                             */
/* Date:        4/10/96                                         */

/* File:        Log.H                                           */
/* Revision:    2.0                                             */

/* System event logger for external tracing                     */

#ifndef __LOG_H__
#define __LOG_H__

/* The trace-file format is very simple at this project stage:  */
/* Each time we create a process we store some useful data into */
/* the descriptor using a set of four different entries; any    */
/* other action is recorded using its current time, the PID of  */
/* the active (EXEC) task and the action/status which the task  */
/* is going to perform. The hooks are embedded directly into    */
/* the kernel primitives and are actually compiled if __LOG__   */
/* is currently #defined!                                       */

#define NAME_LEN 8

/* The following structures needs to be aligned at byte boundary! */
/* Enforce this using proper compiler switch			  */
typedef struct {
	DWORD t;
	BYTE pid;
	BYTE act;
	BYTE st;	
} LOG_DES;

typedef struct {
	DWORD period;
	DWORD deadline;
	DWORD creation;
	BYTE type_class;
	char name[NAME_LEN];
} LOG_DATA;

#define ACT_NIL          	0
#define ACT_CREATE       	1
#define ACT_DISPATCH     	2
#define ACT_SCHEDULE     	3
#define ACT_ACTIVATE     	4
#define ACT_SLEEP        	5
#define ACT_WAKEUP       	6
#define ACT_ENDCYCLE     	7
#define ACT_SIGNAL       	8
#define ACT_WAIT         	9
#define ACT_WAITNB       	9
#define ACT_KILL         	10
#define ACT_DELAY        	11
#define ACT_INTERRUPT		12
#define ACT_START_WINDOW 	13
#define ACT_END_WINDOW   	14
#define ACT_INFO          	15
#define ACT_UPDATE          	16
#define ACT_FIX_FULL        	17
#define ACT_CREATE_NO_TRACE 	18

#define MAX_ACT        		18

#define FASTON_FLAG 	0x40
#define FASTOFF_FLAG 	0x20
#define FAST_FLAG 	0x60

#define MAX_STATE 9

/* $HEADER+     */

/*
    The default size for the log table is 10000 records; only 100
    records are used for the fixed portion! These are only some
    "standard" definition to initialize the LogTable
*/

/* 
    Logging status: they are stored into a control word; the less significant
    bit is set if logging is active; the 2nd bit is set if loop-recording is
    used, cleared if standard recording is used.
    Of course, log can be used just once!
*/

#define LOG_FIX         0x01
#define LOG_LOOP        0x02
#define LOG_STOP        0x04

/*
   The new tracer/logger structure will allow a more flexible data
   recording. In particular, I would like to add events signaling other
   kernel operations, like port & CAB I/O or user level data recording.
   The data recording is organized as follows:
   1) It is possible to record a fixed stream of events
   2) It is possible to record only the last events of a stream
   The buffer is split into 2 sections; the first is handled as a
   cyclic queue, the second one is handled as a standard FIFO list.
   The user can start/stop the recording whenever it wants using
   log_start & log_stop.
   Obviously if the recording is started/stopped at random the generated
   trace could result unreadable: for example if you exclude the recording
   of create/kill action the trace would result unreadable!
*/

/*
   The buf parameter is the pointer to the buffer.
   The splitting parameter marks the length of the cyclic sub-buffer.
   The tb parameter is a reference to the time base of the system.
*/   

/* User level interface toward the tracing support */

/*
   Well, that is not really the best thing to do! We have a portability
   flaw here, due to the necessity of redefining the I/O set of operations
   Anyway, this gives the possibility of setting up different log dump
   methods (i.e. disk, serial or network...)
   The log dump is performed after all task have been stopped but
   before the system has shut down! Well now this is not really true
   because the filesystem is not directly supported and we use DOS
   system call to perform the dump.
   The log dump is open using the referenced function, then the records
   are written to it and it is finally closed. This should be enough
   general... I Hope!
*/

#define FIX_EVENTS    500
#define LOOP_EVENTS   800
#define FIX_WINDOWS     1


typedef struct {
	DWORD fix_events;
	DWORD loop_events;
	BYTE  start_mode;
	BYTE  fix_windows;
	char *name;
} LOG_INFO;

#define BASE_LOG {FIX_EVENTS,LOOP_EVENTS,LOG_FIX,FIX_WINDOWS,"baselog.hrt"}

#define log_set_name(li,n)      	li.name = n;
#define log_set_fix_events(li,x)	li.fix_events = x;
#define log_set_loop_events(li,x)       li.loop_events = x;
#define log_set_start_mode(li,x)	li.start_mode = x;
#define log_set_fix_windows(li,x)       li.fix_windows = x;

/* Init the logger with the parameters encoded into the l structure */
void log_init(LOG_INFO *l);

/* Record a single action; it is the log probe which is called when */
/* a kernel primitive is executed. It is user visible to allow the  */
/* user to monitor its own actions; hovewer this functionality will */
/* be available in the future (Sorry!)...                           */
void log_record(PID p,BYTE act);

/* The logger default function which is called when the system is   */
/* closed to dump the log onto a stable support; at present the     */
/* dump is done on disk, using different tricks for the various     */
/* architectures! I hope the filesystem will come soon...           */
void log_dump(void);

/* These functions are used to change the logging status        */
/* They return the old status value; you can check it using     */
/* boolean operators and LOG_FIX, LOG_LOOP bit mask          */

void log_fix(void);
void log_loop(void);
void log_stop();
void log_reset();


/* The Log system error codes... */

#define LOG_OUT_OF_MEM          121
#define LOG_INIT_ERROR          122
#define LOG_CANNOT_OPEN         123
#define LOG_DUMP_ERROR          124
#define LOG_CANNOT_WRITE        125

/* $HEADER-     */

#endif /* __LOG_H__ */
