/* Project:     HARTIK 3.0                                      */
/* Description: Hard Real TIme Kernel for 8086 compatible       */
/* Author:      Gerardo Lamastra                                */
/*              Giuseppe Lipari                                 */
/* Date:        1/7/96                                          */

/* CVS $Id: sys.h,v 1.5 1999/05/12 07:41:12 massy Exp $ */

/* File:        Sys.H						*/
/* Revision:    3.0                                          	*/
/* Date:	14/3/97						*/

/* Kernel main data structures & primitives			*/
/* $HEADER+							*/

#ifndef __SYS_H__
#define __SYS_H__

#include "sys/const.h" 

/* User level data types & prototypes for system primitives */

#define TASK void __cdecl  /* Used to identify the C-function which */
			   /* are Hartik tasks; it has only a 	*/
			   /* visibility purpose                    */
typedef int QUEUE;
typedef int SEM;
typedef int PID;

/*---------------------------------------------------------------------*/
/* This primitive can only be called from main(); they allow clean     */
/* initialization and exit for HARTIK                                  */
/* Remember that the stack which the HOST-OS passes to the application */
/* is directly used by the main() procedure, which runs as a no-real   */
/* time task; so this stack must be also the hartik exit stack         */
/* If a critical error occurs the system is aborted using the system   */
/* stack which is statically allocated                                 */
/* The task stack is dinamically allocated (using the VM mem manager)  */
/* when the task is created!                                           */
/*---------------------------------------------------------------------*/

/* The system start-up parameters are passed into a structure 	*/
/* This makes it possible to add functionality to the kernel  	*/
/* without the need to change the old applications		*/

typedef struct sys_parms {
	DWORD  tick;
	BYTE   unit;
	double Us;
        int    wcheck;
} SYS_PARMS;

#define BASE_SYS {1000,uSec,0.25,1}

#define sys_def_tick(s,t,u)   s.tick = t; s.unit = u
#define sys_def_Us(s,u)	      s.Us = u
#define sys_def_nocheck(s)    s.wcheck = 0

void sys_init(SYS_PARMS *s);
TIME sys_time(void);
DWORD sys_ticks(void);
DWORD sys_gettick(void);
void sys_end(void);

/* Standard exception handlers */
int sys_abort(int code);
int sys_ignore(int code);

#define ABORT_HANDLER   sys_abort       
#define IGNORE_HANDLER  sys_ignore
#define NULL_HANDLER    NULL

void sys_status(DWORD cw);
/* Control Flag to pass to sys_status */
#define CLOCK_STATUS            0x0001
#define READY_NRT_STATUS	0x0002
#define READY_RT_STATUS		0x0004
#define BLOCKED_STATUS          0x0008
#define IDLE_STATUS             0x0010
#define SLEEP_STATUS            0x0020
#define EXEC_STATUS             0x0040
#define MEM_STATUS              0x0080
#define PORT_STATUS             0x0100

/* Print all ready queues, clock & blocked */
#define NORM_STATUS             0x004F
#define STD_STATUS		0x004F

/* Use this function to post your own exit operations*/
int sys_atexit(void (*exit_code)(void),BYTE when);
/* Status bit values; used to signal that the exit function has to be   */
/* called before/after the kernel has been terminated                   */
#define NONE_EXIT       0
#define BEFORE_EXIT     1
#define AFTER_EXIT      2

/* Task management primitives */

/* Create the task reserving a descriptor */
PID task_create(char *name,TASK (*body)(),BYTE pclass,
		BYTE type,DWORD p,MODEL *m);

/* Kill/Activate a task specified via pid returned from task_create */
int task_kill(PID pid);
int task_activate(PID pid);
					
/* This primitive is used to kill the executing process;               */
/* Its address is pushed onto process stack to prevent the task from   */
/* returning to an undefined point.                                    */
/* Generally the user does not need to use this primitive.             */
void task_abort(void);

/* This function signals to the kernel that the current istance of */
/* the task (periodic or aperiodic) is ended; so the task can be   */
/* suspended until it is activated again.                          */
void task_endcycle(void);

/* This function suspend the actual task until an explicit activation */
void task_sleep(void);
void task_delay(DWORD delay);

/* This primitives refers the group id which is supplied */
/* by the application, not by the kernel                 */
int group_activate(WORD g);
void group_kill(WORD g);

/* These functions return info on a task */
PID task_exec(void);
PID task_pid(char *name);
char *task_name(PID p);
DWORD task_dline(PID p);
DWORD task_priority(PID p);
DWORD task_request(PID p);
BYTE task_status(PID p);

/* User level semaphoric primitives */

SEM  sem_create(int n);			/* Alloc a semaphore descriptor */
void sem_delete(SEM q);			/* Free a    "         "	*/
WORD sem_xwait(SEM s,int n,BYTE wait);	/* Generic wait			*/
void sem_xsignal(SEM s,int n); 		/* Generic signal		*/

#define sem_wait(s,z)	sem_xwait(s,1,z)
#define sem_signal(s)   sem_xsignal(s,1)

/* Interrupt handler installation */
int handler_set(int no, void (*fast)(int), PID pi);
int handler_remove(int no);

#define EXC_CLEAR	0	/* Possible status bit values */
#define EXC_INSTALLED	1
#define EXC_ACTIVE	2
#define EXC_MASKED	4

typedef int (*EXC_HANDLER)();

EXC_HANDLER exc_get(int code);
void exc_set(int code,EXC_HANDLER h);
void exc_raise(int code);

#endif /* __SYS_H__ */

