/* Project: 	HARTIK 3.0      				*/
/* Description: Hard Real TIme Kernel for 8086 compatible	*/
/* Author:	Gerardo Lamastra				*/
/*		Giuseppe Lipari					*/
/* Date:	9/5/96 						*/

/* File: 	VM.H						*/
/* Revision: 	1.0						*/

/* Header file for the Virtual Machine				*/
/* It contains a new type and the prototype of the exported     */
/* functions.							*/
/* The VM layer encloses all the machine features for timer and */
/* for multitasking, so that the kernel is machine independent  */
/* The kernel must respect some constraints about the interface */
/* with the VM. Check the documentation about VM layer		*/
/* These headers enclose every architecture dependent setting   */
/* Using conditional compilation we are able to #include the    */
/* correct one							*/

#ifndef __VM_H__
#define __VM_H__

#ifndef __CONFIG_H__
#include "config.h"
#endif

#ifndef __HW_DEP_H__
#include "hw_dep.h"
#endif

/* Initialize the context for a task. The first parameter is the function */
/* address (starting point of the task), the second is the address of the */
/* stack for that task. The VM DO NOT allocate memory for the stack, so   */
/* is the kernel that reserve this memory.				  */  

CONTEXT __cdecl VM_context_create(void __cdecl (*task)(),BYTE *stack,
				WORD parm,void (*killer)(void),WORD ctrl);

/* Release a used task context						  */

void __cdecl VM_context_delete(CONTEXT c);

/* Put the context value into human readable form; used for debug! */

char *VM_context_sprintf(char *str,CONTEXT c);

/* Output function for kernel */
/* Mainly used during debugging */

int VM_printf(char *fmt,...);

/* These 3 function realize the context switching. The context_save has   */
/* to be the first call of a kernel primitive, and the context_change has */
/* to be the last call.							  */
/* The context_save disables the interrupt (a kernel primitive must be	  */
/* atomic) and return the context of the running task.			  */
/* The context_change take the context of the new task (or of the	  */
/* same task), switch to the new context and return restoring the flag	  */
/* register.								  */
/* The context_load is used when the task is going to be killed; then its */
/* context does not need to be saved; we only need to load the context of */
/* the new task; the effective implementation of this functions can vary  */
/* greatly throughout different implementations as some of them are       */
/* mapped to empty functions or mapped one onto another			  */

CONTEXT __cdecl VM_context_save(void);
void    __cdecl VM_context_change(CONTEXT c);
void    __cdecl VM_context_load(CONTEXT c);

/* These functions start-up & close the VM layer */

void VM_init(TIME value);
void VM_end(void);

/* These functions are used to mask/unmask selectively interrupts    */
/* The irq services are also #defined to allow more generic inteface */
/* This is done into hw... files!				     */

void VM_irq_mask(WORD irqno);
void VM_irq_unmask(WORD irqno);

/* These functions provide direct access to interrupt table 	*/
/* We can write the HARTIK interrupt table but only read	*/
/* the host OS interrupt table!					*/
void VM_irq_set(WORD irqno,INTERRUPT handler);
INTERRUPT VM_irq_get(WORD irqno);

/* This functions acts as safety place where to go when any error */
/* occurs and we do not know what context is active		  */

void __cdecl VM_abort(int code);

/* This function is used to read the status of the VM timer 	*/
/* It is used to map uniformly the access to timer chip between */
/* the various architectures supported				*/
TIME VM_timer_read(void);

/* This function is used to switch the VM_timer_read handler	 */
/* For example, under P5 or AXP it is possible to use the free	 */
/* running counter of the CPU; in general it is possible to have */
/* an external timer source, for example from an external board  */
void VM_timer_set_handler(TIME (*h)(void));

/* These function are used to manage memory at VM layer */

void * VM_alloc(DWORD size);
WORD VM_free(void *ptr,DWORD size);
void VM_mem_init(void *base,DWORD size);
void VM_mem_dump(void);

/* The following are kernel primitive called from the interrupt		*/
/* routine of the VM.							*/
/* The wake_up() is the handler of the timer : it receives the active	*/
/* context and return a new context (that's because it can operate	*/
/* a preemption). It can't call the context_save or the context_change,	*/
/* nor can enable interrupt.						*/
/* The act_int() is the generic handler for an interrupt: it receives	*/
/* the number of the interrupt, and the active context, and return	*/
/* the new context.Like the wake_up(), it cannot call the context_save	*/
/* or the context_change, but it can enable interrupt.			*/

CONTEXT	__cdecl wake_up(CONTEXT c);
CONTEXT __cdecl act_int(CONTEXT c, BYTE n);

/* Well, is sounds strange to have these functions here; they are used  */
/* to allow math calculation into a kernel primitive!			*/
void VM_FPU_save(void);
void VM_FPU_restore(void);

/* The following structure is filled up by the VM_init() and can be 	*/
/* read by the kernel modules to know about the architecture		*/
/* Each time a new hardware CPU is added to the VM layer, you have to	*/
/* define a new XXX_ARCH structure, whose first field is a char *arch	*/
/* This is used to identify the architecture, then subsequent field	*/
/* can be decoded!!							*/

/* WARNING: I tried to use bitfields, but this caused am INTO error?!?  */
/* only when using GNU-C!! So i preferred to use standard DWORD flags	*/

/* Capabilities */
#define VM_X86_HAS_INVLPG 	0x01
#define VM_X86_HAS_CPUID	0x02
#define VM_X86_HAS_FPU		0x04
#define VM_X86_INTERNAL_FPU	0x08
#define VM_X86_HAS_TSTAMP	0x10

/* Bugs */
#define VM_X86_FDIV_BUG		0x01
#define VM_X86_F00F_BUG		0x02

typedef struct {
    char *arch;
    int cpu;	/* 0,1,2,3,4,5,6 -> 8086/8,80186,80286,80386,80486,P5,PII o Overdrive */
    int fpu;	/* 0,1,2,3 -> None,8087,80287,80387 */
    char *model; /* Dx, Dx2, ... */
    char *vendor; /* Intel, Cyrix, AMD or unknown */
    DWORD capabilities;
    DWORD bugs;
    /* BUGs!! Warning: Currently, no workaround is available!
    */
    int f00f_bug;
    int fdiv_bug;
} X86_ARCH;

typedef struct {
    char *arch;
    /* Tonino, fill up this stuff! */
} AXP_ARCH;

typedef union {
    char *arch;
    X86_ARCH x86;
    AXP_ARCH axp;
} VM_ARCH;

extern VM_ARCH VM_arch;

#endif /* __VM_H__ */
