/* Project:     HARTIK 3.0                                      */
/* Description: Hard Real TIme Kernel for 8086 compatible       */
/* Date:        2/7/96                                          */

/* File:           Cabs.C                                       */
/* Translated by : Giuseppe Lipari                              */
/* Revision:       1.1                                          */

#include <string.h>
#include <stdlib.h>

#include "vm.h"
#include "kern.h"
#include "cabs.h"

#define MAX_NAME 10

/* "cab":       contiene dei buffer per memorizzare messaggi            */
/* "cab_data":  buffer contenente il messaggio e informazioni relative  */
/* "messaggio": si trova a partire da (*cab_data + 1)                   */

struct cab_data {		/* struttura del buffer di cab  */
    struct cab_data *next;	/* successivo buffer del cab    */
    unsigned short n_uso;	/* processi che usano il buffer */
};

struct cab_desc {			/* struttura descrittore di cab */
	char    name[MAX_NAME+1];	/* nome del CAB                 */
	CAB     next_cab_free;		/* indice successivo cab libero */
	BYTE    busy;			/* cab libero/occcupato         */
	char    *mem_cab;		/* memoria globale cab          */
	BYTE    n_buf;			/* numero dei buffer nel cab    */
	BYTE    nfree;			/* numero buffer liberi         */
	unsigned dim_mes;		/* dimensione del messaggio     */
	struct  cab_data *free;		/* puntatore primo buffer libero*/
	struct  cab_data *mrd;		/* puntatore must_recent_data   */
};

static char *cab_error_msg[] = {
    "Unvalid number of message","No more cab free",
    "Number of message exceeded","Cab id not valid",
    "Cab is closed"};

static struct   cab_desc  cabs[MAX_CAB]; /* vettore descrittori dei CAB  */
static CAB      free_cab;                /* indice del primo cab libero  */

static int checkcab(CAB id)
{
    if (id >= MAX_CAB) {
	proc_table[exec].errno = CAB_UNVALID_ID;
	return -1;
    }
    if (cabs[id].busy == TRUE) return TRUE;
    else proc_table[exec].errno = CAB_CLOSED;
    return -1;
}

static int cab_exception(int code)
{
    VM_printf("\nCab exception\n");
    VM_printf("Code [%d]\nCause : %s\n",code,cab_error_msg[code-CAB_UNVALID_MSG_NUM]);
    /* Cab default action is an abort */
    return(0);
}

/*----------------------------------------------------------------------*/
/* cab_init -- inizializza le strutture dei cab                         */
/*----------------------------------------------------------------------*/
void cab_init(void)
{
    int i;

    free_cab = 0;
    for (i=0; i < MAX_CAB - 1; i++) {
	cabs[i].next_cab_free = i+1;
	cabs[i].busy = FALSE;
    }
    cabs[MAX_CAB-1].next_cab_free = NIL;
    cabs[MAX_CAB-1].busy = FALSE;
    for (i = CAB_UNVALID_MSG_NUM; i <= CAB_CLOSED; i++)
	exc_set(i,cab_exception);
}

/*----------------------------------------------------------------------*/
/* cab_create -- crea un cab, lo inizializza e restituisce l'indice     */
/*----------------------------------------------------------------------*/
CAB cab_create(char *name, int dim_mes, BYTE num_mes)
{
    CAB id;                     /* indice del cab da restituire         */
    struct cab_desc *pid;       /* puntatore al cab (velocizza accesso) */
    char *mem;                  /* puntatore di appoggio al buffer      */
    struct cab_data *tmp;       /* puntatore di scorrimento lista cab   */
    int i;                      /* variabile indice                     */
    
    VM_disable();

    /* Se non ci sono piu' cab liberi o il parametro num_mes < 1 */
    /* solleva l'eccezioni                                       */

    if (num_mes < 1) {
	proc_table[exec].errno = CAB_UNVALID_MSG_NUM;
	VM_enable();
	return -1;
    }
    if ((id=free_cab) != MAX_CAB) {
	pid = &cabs[id];        /* prendo l'indirizzo del cab */
	free_cab = pid->next_cab_free;
    }
    else {
	proc_table[exec].errno = CAB_NO_MORE_ENTRY;
	VM_enable();
	return -1;
    }

    /* richiede un identificatore e la memoria */
    mem = VM_alloc((dim_mes + sizeof(struct cab_data)) * num_mes);

    VM_enable();
    /* inizializzazione del descrittore del cab */

    strcpy(pid->name, name);
    pid->mem_cab = mem;
    pid->dim_mes = dim_mes;
    pid->n_buf = num_mes;

    /* inizializzazione primo messaggio e buffer liberi */

    pid->mrd = (struct cab_data *)mem;
    i = (int)num_mes;
	tmp = NULL;
    while (i--) {
	tmp = (struct cab_data *)mem;
	mem += sizeof(struct cab_data) + dim_mes;
	tmp->next = (struct cab_data *)mem;
	tmp->n_uso = 0;
    }

    tmp->next = NULL;
    pid->free = pid->mrd->next;

    mem = (char *)(pid->mrd + 1);
    for (i=0; i<dim_mes; i++) *(mem++) = 0;
    pid->nfree = num_mes - 1;

    VM_disable();
    pid->busy = TRUE;
    VM_enable();

    return(id);
}

/*----------------------------------------------------------------------*/
/* cab_reserve --  richiede un buffer in cui mettere i dati da inviare  */
/*              ritorna un puntatore al buffer                          */
/*----------------------------------------------------------------------*/
char *cab_reserve(CAB id)
{
    struct cab_desc *pid;
    char *buf;

    /* controlla l'identificatore del CAB */
    if (checkcab(id) == -1) return(NULL);

    pid = &cabs[id];
    VM_disable();
    /* Se il numero di elementi assegnati non e` = al massimo */
    if ((pid->nfree)--) {
	buf = (char *)(pid->free + 1);
	pid->free = pid->free->next;
	VM_enable();
	return(buf);
    }
    else {
	proc_table[exec].errno = CAB_TOO_MUCH_MSG;
	VM_enable();
	return(NULL);
    }
}

/*----------------------------------------------------------------------*/
/*  cab_putmes -- immette un nuovo messaggio nel cab                    */
/*----------------------------------------------------------------------*/
int cab_putmes(CAB id, char *pbuf)
{
    struct cab_data *pold;
    struct cab_desc *pid;

    if (checkcab(id) == -1) return -1;
    pid = &cabs[id];

    VM_disable();
    pold = pid->mrd;
    if (pold->n_uso == 0) {
	pold->next = pid->free;
	pid->free = pold;
	(pid->nfree)++;
    }

    pid->mrd = ((struct cab_data *)pbuf) - 1;
    VM_enable();
    return 1;
}

/*----------------------------------------------------------------------*/
/*  cab_getmes  --  richiede l'ultimo messaggio presente nel cab.       */
/*               Ritorna un puntatore al buffer piu' recente            */
/*----------------------------------------------------------------------*/
char *cab_getmes(CAB id)
{
    char *tmp;

    if (checkcab(id) == -1) return(NULL);
    VM_disable();

    /* cabs[id].mrd punta all'ultimo buffer inserito, incremento        */
    /* il puntatore di uno e ottengo l'area del messaggio, converto     */
    /* quindi il puntatore al tipo carattere. Segue l'incremento        */
    /* del campo contatore di uso buffer                                */

    (cabs[id].mrd->n_uso)++;
    tmp = (char *)(cabs[id].mrd + 1);

    VM_enable();
    return(tmp);
}

/*----------------------------------------------------------------------*/
/* cab_unget -- segnala che il task non usa piu' il messaggio,          */
/*              se questo non e' piu' usato da nessuno viene rilasciato */
/*              ritorna un risultato                                    */
/*----------------------------------------------------------------------*/
int cab_unget(CAB id, char *pun_mes)
/* small id;                            indice del cab di lavoro        */
/* char *pun_mes;                       puntatore al messaggio          */
{
    struct cab_data *pbuf;
    struct cab_desc *pid;

    if (checkcab(id) == -1) return -1;
    pid = &cabs[id];

    VM_disable();
    pbuf = ((struct cab_data *)(pun_mes)) - 1;

    if ((--(pbuf->n_uso) == 0) && (pbuf != pid->mrd)) {
	pbuf->next = pid->free;
	pid->free = pbuf;
	(pid->nfree)++;
    }
    VM_enable();
    return 1;
}

/*----------------------------------------------------------------------*/
/*      cab_delete -- libera la memoria e rilascia il cab               */
/*----------------------------------------------------------------------*/
void cab_delete(CAB id)
{
    struct cab_desc *pid;

    pid = &cabs[id];
    VM_disable();
    VM_free(pid->mem_cab,(pid->dim_mes + sizeof(struct cab_data))*pid->n_buf);

    pid->busy = FALSE;
    pid->next_cab_free = free_cab;
    free_cab = id;
    VM_enable();
}
