/* Project:     HARTIK 3.0                                      */
/* Description: Hard Real TIme Kernel for 8086 compatible       */
/* Author:      Gerardo Lamastra                                */
/* Date:        1/10/96                                         */

/* File:        Exc.C                                           */
/* Revision:    1.1                                             */

/* This module contains all the stuff necessary to handle       */
/* interrupts & exceptions                                      */

#include <string.h>
#include <stdlib.h>

#include "vm.h"
#include "kern.h"

#ifdef __TRACE__
    #include "log.h"
#endif

/* Interrupt handling table */
static struct int_des {
    void (*fast)(int n);
    PID proc_index;
    BYTE isActive, isUsed;
} int_table[MAX_INT];

/* Exception handling table */
static struct exc_des {
    EXC_HANDLER h;
    BYTE status;
} exc_table[MAX_EXC];

/* Indicates if there is an active fast routine for interrupt */
BYTE isActiveInt = FALSE;

extern DWORD last_int;	/* Time of the last system call		*
			* (in usec) --- Luca 			*/
extern DWORD act_time;	/* Actual time, readed by every system	  *
			 * call that manages the capacity...	  *
			 * It's global because the task_makeready *
			 * problem...				  */

/* Used only for  DEBUG...*/
extern int ldbg;

/* The act_int & act_exc are the high level kernel entry point  */
/* to interrupt management & exception handling; Warning the    */
/* interrupt can cause a preemption!                            */
/* The fast handler is a standard piece of code which runs with */
/* interrupts enabled to allow interrupt nesting                */

CONTEXT __cdecl act_int(CONTEXT c,BYTE n)
{
    extern void task_makeready(PID p);
    register unsigned i = 0, intAct = 0;
    PID p;
    
    /* First, we have to manage the capacity of the running task...	*/
    /* 		----By Luca	:-(					*/
    CBS_PROLOGUE();

    while ((i < MAX_INT) && (intAct == 0)) {
	if (int_table[i].isActive == TRUE) intAct++;
	i++;
    }
    /* Set the Interrupt Active Flag; Only the last interrupt   */
    /* can perform the real scheduling!                         */
    isActiveInt = TRUE;
    /* If a fast handler is linked, call it                     */
    /* Ehy Remember this!                                       */
    /* A fast handler is a STANDARD piece of application code   */
    /* then any kernel primitive can be invoked from within!    */
    /* You must pay a lot of attention to this because context  */
    /* handling functions can make some mess.                   */
    /* To understand this problem, check the VM1.ASM module     */
    /* in the 16 bit kernel version!                            */

    #ifdef __TRACE__
	log_record(exec,FASTON_FLAG | n);
    #endif    
    int_table[n].isActive = TRUE;
    if (int_table[n].fast != NULL) {
	VM_enable();
	(int_table[n].fast)(n);
	VM_disable();
    }
    int_table[n].isActive = FALSE;
    #ifdef __TRACE__
	log_record(exec,FASTOFF_FLAG | n);
    #endif    

    /* If a sporadic process is linked,activate it */
    p = int_table[n].proc_index;
    if (p != NIL) task_makeready(p);
    #ifdef __TRACE__
	log_record(p,ACT_INTERRUPT);
    #endif
    
    /* If there isn't any other active interrupt, save the context */
    /* & call the scheduler, then make the context change          */
    if (intAct == 0) {
	proc_table[exec].context = c;
	isActiveInt = FALSE;
	scheduler();
	c = proc_table[exec].context;
    }
    /* Else return the actual context */
    return c;
}

/* Whenever an exception of any kind is detected, the control is passed */
/* to the following procedure which calls the appropriate handler       */
/* If the exception is reiterated an abort is issued immediately        */

void __cdecl act_exc(int n)
{
    int result;
    if (exc_table[n].status & EXC_INSTALLED) {
	if (exc_table[n].status & EXC_ACTIVE) {
	    sys_abort(DOUBLE_EXCEPTION);
	} else {
	    exc_table[n].status |= EXC_ACTIVE;
	    result = exc_table[n].h(n);
	    exc_table[n].status &= ~EXC_ACTIVE;
	    if (result) return;
	    else sys_abort(n);
	}
    }
}
    
void excirq_init(void)
{
    register i;

    for (i = 0; i < MAX_INT; i++) {
	int_table[i].fast = NULL;
	int_table[i].proc_index = NIL;
	int_table[i].isActive = FALSE;
	int_table[i].isUsed = FALSE;
    }
    /* Init exception table */
    /* Exceptions are grouped into blocks; the first represents */
    /* the VM exceptions as division by 0 or float-math error;  */
    /* the second block represents kernel errors such as time   */
    /* overflow; finally there are the primitive hooks, which   */
    /* are issued whenever a primitive is invoked; they are     */
    /* useful for monitoring/debugging.                         */
    for (i = 0; i < 64; i++) {
	exc_table[i].h = IGNORE_HANDLER;
	exc_table[i].status = EXC_INSTALLED;
    }
    for (i = 64; i < 128; i++) {
	exc_table[i].h = ABORT_HANDLER;
	exc_table[i].status = EXC_INSTALLED;
    }
    for (i = 128; i < MAX_EXC; i++) {
	exc_table[i].h = NULL_HANDLER;
	exc_table[i].status = EXC_CLEAR;
    }
}

/* For debug only! Print the Interrupt table entries */
void dump_irq(void)
{
    register i;
    for (i = 0; i < 16; i++) {
	VM_printf("Irq[%d] Fast : %p Pid : %d\n",i,int_table[i].fast,int_table[i].proc_index);
    }
}

/*----------------------------------------------------------------------*/
/* Interrupt table management. The following function install the fast  */
/* handler and the sporadic task linked to the interrupt no.            */
/* If the fast parameter is NULL, no handler is called.                 */
/* If the pi parameter is NIL no task is installed                      */
/*----------------------------------------------------------------------*/
int handler_set(int no, void (*fast)(int n), PID pi)
{
    SYS_FLAGS f;
         
    if ((no < 1) || (no > 15)) {
			proc_table[exec].errno = WRONG_INT_NO;
			return -1;
	}
    f = VM_save();
    if (int_table[no].isUsed == TRUE) {
			VM_restore(f);
			proc_table[exec].errno = USED_INT_NO;
			return -1;
	}
    int_table[no].fast = fast;
    int_table[no].proc_index = pi;
    int_table[no].isActive = FALSE;
    int_table[no].isUsed = TRUE;
    VM_irq_unmask(no);
    VM_restore(f);
	return 1;
}

int handler_remove(int no)
{
    SYS_FLAGS f;
         
    if (no < 1 || no > 15) {
			proc_table[exec].errno = WRONG_INT_NO;
			return -1;
	}
    f = VM_save();
    if (int_table[no].isUsed == FALSE) {
			VM_restore(f);
			proc_table[exec].errno = UNUSED_INT_NO;
			return -1;
	}
    int_table[no].fast = NULL;
    int_table[no].proc_index = NIL;
    int_table[no].isUsed = FALSE;
    VM_irq_mask(no);
    VM_restore(f);
	return 1;
}

/*----------------------------------------------------------------------*/
/* Exception handling functions used to get/set/call an entry into      */
/* the exception table                                                  */
/*----------------------------------------------------------------------*/
EXC_HANDLER exc_get(int code)
{
    return(exc_table[code].h);
}

void exc_set(int code,EXC_HANDLER h)
{
    SYS_FLAGS f;

    f = VM_save();
    if (h == NULL) exc_table[code].status = 0;
    else {
	exc_table[code].h = h;
	exc_table[code].status = EXC_INSTALLED;
    }
    VM_restore(f);
}

void exc_raise(int code)
{
    SYS_FLAGS f;

    f = VM_save();
    act_exc(code);
    VM_restore(f);
}

