/* Project:     HARTIK 3.0                                      */
/* Description: Hard Real TIme Kernel for 8086 compatible       */
/* Author:      Gerardo Lamastra                                */
/*              Giuseppe Lipari                                 */
/* Date:        1/10/96                                         */

/* CVS $Id: kern.c,v 1.14 1999/11/04 10:15:19 massy0 Exp $ */

/* File:        Kern.C                                          */
/* Revision:    3.1.0                                           */

/* This file contains all the kernel level functions and most   */
/* of basic task primitives.                                    */

/* changelog:
 * -- added __SAFEXBIOS__ initialization (by Massy)
 * -- added: can be passed NULL at sys_init()
 */

#include <string.h>
#include <mem.h>
#include <stdlib.h>

#include "vm.h"
#include "kern.h"

#ifdef __PORT__
    #include "port.h"
    extern void port_init(void);
#endif

#ifdef __CAB__
    #include "cabs.h"
    extern void cab_init(void);
#endif

#ifdef __TRACE__
    #include "log.h"
#endif    

#ifdef __SRP__
    #include "srp.h"
    #define SRP_CHECK()     (proc_table[ready_rt].period < sys_ceiling)
#else
    /* If SRP isn't used they are always true */
    #define SRP_CHECK()             1
    #define srp_stack_empty()       1
    void srp_endtask(PID p)
    {
	proc_table[p].status = FREE; 
	q_insert(p,&freedesc);       
    }
#endif

#ifdef __SAFEXBIOS__
    #include "x86/xsys.h"
#endif

void _endtask(PID p)
{

#ifdef __SRP__
  if (proc_table[p].pclass == HARD) {
    srp_endtask(p);
    return;
  }
#endif
  
  proc_table[p].status = FREE; 
  q_insert(p,&freedesc);       

}

/*----------------------------------------------------------------------*/
/* Kernel System variables                                              */
/*----------------------------------------------------------------------*/
int wcheck;              /* WCET Monitoring on/off */
PID exec;               /* Currently executing process          */

QUEUE freedesc;         /* Free descriptor handled as a queue   */
QUEUE ready_rt;         /* Ready queue for real-time tasks      */
QUEUE ready_nrt;        /* Ready queue for the others           */
QUEUE idle;             /* Idle periodic task queues            */
QUEUE zombie;           /* Zombie queue                         */

DWORD sys_clock = 0;    /* System clock (in tick numbers)       */
DWORD sys_tick;         /* System tick (in usec)                */
DWORD sys_flags;	/* System status flags			*/
			/* Now, just SYS_AUTOABORT is defined	*/

DWORD last_int = 0;		/* Time of the last system call
			   (in usec) --- Luca 			*/
DWORD act_time = 0;		/* Actual time, readed by every system
			   call that manages the capacity...
			   It's global because the task_makeready
			   problem...				*/

			   /* 4 DBG...*/ int ldbg=0;

/* Process descriptor table */
struct proc_des proc_table[MAX_PROC];

/* List of function to call upon exit; they are posted with sys_atexit  */
/* They can be called before/after the VM layer has been closed         */
static struct exit_func {
    void (*exit_code)(void);
    BYTE status;
} exit_list[MAX_EXITFUNC];


/* Kernel error messages */
static char *kern_error_msg[] = {
    "Clock Overrun","Hard deadline missed",
    "No more task descriptor available",
    "No more memory available for stack allocation",
    "Invalid Task_set.. operation\nTask not sleeping",
    "Wrong queue","Semaphore unvalid","No more semaphore left",
    "EXC: Wrong interrupt mask/unmask","Invalid killing",
    "Invalid create","Invalid group id (0)","Out of memory",
    "SRP: Cannot guarantee hard task",
    "SRP: Stack is full","SRP: stack is empty",
    "SRP: Tried to extract a non used entry",
    "Too much posted exit functions",
    "WCET violation",
    "Invalid Task ID",
	"EXC: request a busy irq",
	"EXC: trying to free a non-used irq",
	"Unknown error"};

    
/*----------------------------------------------------------------------*/
/* Kernel internal functions                                            */
/*----------------------------------------------------------------------*/

/* The main must be linked to a descriptor in order to preemp it        */
/* The main will inherith the context assigned to the entire application*/
/* from the host OS                                                     */
extern void main();

/* Dummy process declaration */
static TASK dummy(void);

/* Dummy task must be present & cannot be killed;                       */
/* Its stack is statically allocated                                    */
static BYTE dummy_stack[STACK_SIZE];

static TASK dummy()
{
    /*
    It is possible to Halt the CPU & avoid consumption if idle
    cycle are intercepted with hlt instructions!
    It seems that some CPU have buggy hlt instruction or they
    have not is at all! So, if available, use the hlt facility!!
    */
    #ifdef __HLT_WORKS__
    FOREVER {
	hlt();
    }
    #else
    FOREVER;
    #endif
}

/* The wake_up & act_int are the kernel interface toward the VM         */
/* The argument is the context of the actual task, while the result     */
/* is the context of the new task, which could be different from the    */
/* last executing because of preemption.                                */
/* We choose to implement this ones following standard C conventions    */
CONTEXT __cdecl wake_up(CONTEXT c)
{
    PID p;
    WORD count = 0;

    /* Increment system clock                   */
    sys_clock++;

    /* First, we have to manage the capacity of the running task...	*/
    /* 		----By Luca	:-(					*/
    {
	act_time = sys_clock * sys_tick;
	proc_table[exec].cap -= (act_time - last_int);
	last_int = act_time;		/* update the last int time */
    }

    /* Check for hard tasks overrunning         */
    if (((proc_table[exec].pclass == HARD) || (proc_table[exec].pclass == SOFT)) &&
	(sys_clock > proc_table[exec].dline)) act_exc(HARD_MISS);
    if (!EMPTY(blocked) && (proc_table[blocked].pclass == HARD) &&
	(sys_clock > proc_table[blocked].dline)) act_exc(HARD_MISS);

    /* Remove the zombie task which have passed period end */
    while (!EMPTY(zombie) && (proc_table[zombie].priority <= sys_clock)) {
	p = q_getfirst(&zombie);
       _endtask(p);
    }

    /* Make any periodic task switch */
    while (!EMPTY(idle) && (proc_table[idle].priority <= sys_clock)) {
	p = q_getfirst(&idle);
	proc_table[p].priority = proc_table[p].dline;
	proc_table[p].status = READY;
	/* When a new CBS task arrives to the system, we must check the
	   residual CPU band of the CBS... --- Luca */
	if ((proc_table[p].pclass == SOFT) && (proc_table[p].dline <
		proc_table[p].request + ((proc_table[p].cap * proc_table[p].period) / proc_table[p].q))){
		    proc_table[p].cap = proc_table[p].q;
		    proc_table[p].dline = proc_table[p].request + proc_table[p].period;
		    proc_table[p].priority = proc_table[p].request + proc_table[p].period;
	}
	/* CBS tasks also must be scheduled in rt_queue --- Luca */
	if (proc_table[p].pclass == HARD || proc_table[p].pclass == SOFT) q_insert(p, &ready_rt);
	else q_insert(p, &ready_nrt);
	#ifdef __TRACE__
	    log_record(p,ACT_WAKEUP);
	#endif
	count++;
    }

    if (!isActiveInt) {
	/* Store the context status into the task descriptor only if */
	/* there isn't any active interrupt                          */
	proc_table[exec].context = c;
	scheduler();
	c = proc_table[exec].context;
    }
    return c;
}

/* This function (improperly called scheduler) checks if preemption is  */
/* necessary  and makes it; It has to be called every time a task       */
/* goes in the ready queue                                              */
void scheduler(void)
{
    PID lastexec;
    int count;

    if (isActiveInt || (proc_table[exec].control & NO_PREEMPT)) return;

    if (proc_table[exec].pclass == HARD || proc_table[exec].pclass == SOFT) {
	if (!EMPTY(ready_rt) &&
	    (proc_table[ready_rt].priority < proc_table[exec].priority)
	    && SRP_CHECK()) {
		q_insert(exec,&ready_rt);
		proc_table[exec].status = READY;
		#ifdef __TRACE__
		    log_record(exec,ACT_SCHEDULE);
		#endif
		exec = q_getfirst(&ready_rt);
		proc_table[exec].status = EXE;
		#ifdef __TRACE__
		    log_record(exec,ACT_DISPATCH);
		#endif
	}
    }
    else {
	if (!EMPTY(ready_rt)) {
	    q_insert(exec,&ready_nrt);
	    proc_table[exec].status = READY;
	    #ifdef __TRACE__
		log_record(exec,ACT_SCHEDULE);
	    #endif
	    exec = q_getfirst(&ready_rt);
	    proc_table[exec].status = EXE;
	    #ifdef __TRACE__
		log_record(exec,ACT_DISPATCH);
	    #endif
	}
	else if (proc_table[ready_nrt].priority < proc_table[exec].priority) {
	    q_insert(exec,&ready_nrt);
	    proc_table[exec].status = READY;
	    #ifdef __TRACE__
		log_record(exec,ACT_SCHEDULE);
	    #endif
	    exec = q_getfirst(&ready_nrt);
	    proc_table[exec].status = EXE;
	    #ifdef __TRACE__
		log_record(exec,ACT_DISPATCH);
	    #endif
	}
    }

    lastexec = -1;
    while (lastexec != exec) {
	if (proc_table[exec].pclass == SOFT) {
	    lastexec = exec;
	    count = 0;
	    /* If the task is scheduled by CBS, check for exausted capacity */
	    while (proc_table[exec].cap < (sys_tick - (act_time % sys_tick)) ) {
		proc_table[exec].cap += proc_table[exec].q;
		proc_table[exec].dline += proc_table[exec].period;
		proc_table[exec].priority += proc_table[exec].period;
		count++;
		if (count > MIN_Q) {
		    sys_abort(56000);
		}
	    }

	    /* Call the scheduler & switch task, or return the last context */
	    if (count > 0) {
		if (!EMPTY(ready_rt) &&
		  (proc_table[ready_rt].priority < proc_table[exec].priority)
	    	  && SRP_CHECK()) {
		      q_insert(exec,&ready_rt);
		      proc_table[exec].status = READY;
		      #ifdef __TRACE__
		      log_record(exec,ACT_SCHEDULE);
		      #endif
		      exec = q_getfirst(&ready_rt);
		      proc_table[exec].status = EXE;
		      #ifdef __TRACE__
		      log_record(exec,ACT_DISPATCH);
		      #endif
		  }
	    }
	} else if (proc_table[exec].pclass == HARD) {
	    /* Here, there must be the WCET check... see budget management */
	    if ((sys_tick - (act_time % sys_tick)) - proc_table[exec].cap > proc_table[exec].wcet)
	    {
              if (wcheck) {
                    act_exc(WCET_VIOLATION);
                }
	    }
	    lastexec = exec; /* only to make the loop finish... */
	} else {
	    lastexec = exec; /* only to make the loop finish... */
	}
    }
}

/* This function put a task in exec. It has to be called every time a   */
/* task is descheduled                                                  */
void dispatch(void)
{
    PID lastexec;
    int count;

    if (!EMPTY(ready_rt)) exec = q_getfirst(&ready_rt);
    else exec = q_getfirst(&ready_nrt);
    proc_table[exec].status = EXE;
    #ifdef __TRACE__
	log_record(exec,ACT_DISPATCH);
    #endif

    lastexec = -1;
    while (lastexec != exec) {
	if (proc_table[exec].pclass == SOFT) {
	    lastexec = exec;
	    count = 0;
	    /* If the task is scheduled by CBS, check for exausted capacity */
	    while (proc_table[exec].cap < (sys_tick - (act_time % sys_tick)) ) {
		proc_table[exec].cap += proc_table[exec].q;
		proc_table[exec].dline += proc_table[exec].period;
		proc_table[exec].priority += proc_table[exec].period;
		count++;
		if (count > MIN_Q) sys_abort(56001);
	    }

	    /* Call the scheduler & switch task, or return the last context */
	    if (count > 0) {
		if (!EMPTY(ready_rt) &&
		  (proc_table[ready_rt].priority < proc_table[exec].priority)
	    	  && SRP_CHECK()) {
		      q_insert(exec,&ready_rt);
		      proc_table[exec].status = READY;
		      #ifdef __TRACE__
		      log_record(exec,ACT_SCHEDULE);
		      #endif
		      exec = q_getfirst(&ready_rt);
		      proc_table[exec].status = EXE;
		      #ifdef __TRACE__
		      log_record(exec,ACT_DISPATCH);
		      #endif
		  }
	    }
	} else if (proc_table[exec].pclass == HARD) {
	    /* Here, there must be the WCET check... see budget management */
	    lastexec = exec; /* only to make the loop finish... */
	    if ((sys_tick - (act_time % sys_tick)) - proc_table[exec].cap > proc_table[exec].wcet)
	    {
                if (wcheck) {
                    act_exc(WCET_VIOLATION);
                }
            }
	} else {
	    lastexec = exec; /* only to make the loop finish... */
	}
    }
}

/*----------------------------------------------------------------------*/
/* Kernel main system functions                                         */
/*----------------------------------------------------------------------*/

/* This function initialize the virtual machine (timer, interrupt, mem) */
/* the system's structures (queues, tables) , & the two task main &     */
/* dummy, that are always present                                       */
void sys_init(SYS_PARMS *s)
{
    static SYS_PARMS defaultparms=BASE_SYS;
    extern void C8042_restore(void);
    int i;
    CONTEXT c;

    if (s==NULL) s=&defaultparms;
   
    /* The kernel startup MUST proceed with int disabled!    */
    VM_disable();

    /* Store the TBS utilization factor in integral form */
/*    TBS_util = (DWORD)(s->Us * TBS_PREC);*/
    wcheck = s->wcheck;

    /* Clear the task descriptors */
    for (i = 0; i < MAX_PROC; i++) proc_table[i].status = FREE;
    /* Set up the varius queues & the exec pointer      */
    exec = MAIN_INDEX;
    freedesc = 0;
    ready_rt = NIL;
    ready_nrt = NIL;
    idle = NIL;
    zombie = NIL;
    for (i = 0; i < MAX_PROC-1; i++) proc_table[i].next = i+1;
    proc_table[MAX_PROC-1].next = NIL;
    for (i = MAX_PROC-1; i > 0; i--) proc_table[i].prev = i-1;
    proc_table[0].prev = NIL;
    for (i = 0; i < MAX_EXITFUNC; i++) exit_list[i].status = NONE_EXIT;
    /* Init semphores                                   */
    sem_init();
    /* Init the exception & interrupt tables            */
    excirq_init();
    /* Init the SRP protocol structures */
    #ifdef __SRP__
	srp_init();
    #endif
    
    #ifdef __PORT__
	/* Init the port's stuctures only if port.c is linked */
	port_init();
    #endif

    #ifdef __CAB__
	/* Init the cab's stuctures only if cab.c is linked */
	cab_init();
    #endif

    #ifdef __SAFEXBIOS__
        xbios_initfun();
    #endif
   
    /* Init VM layer (Interrupts & memory management)           */
    /* This MUST be done before we call the VM_context_create() */
    if (s->unit == mSec) s->tick *= 1000;

    /* test on system tick */
    if (s->tick>=55000)  {
       VM_printf("The system tick must be less than 55 mSec!");
       exit(0);
    }
   
    /* Store system time base */
    sys_tick = s->tick;
    VM_init(sys_tick);
    
    /* Fill the main() descriptor                       */
    /* We don't need to initialize main context;        */
    /* This will be done at first task switch           */
    proc_table[MAIN_INDEX].stack = NULL;
    proc_table[MAIN_INDEX].body = (TASK *)(main);
    strcpy(proc_table[MAIN_INDEX].name,"Main");
    proc_table[MAIN_INDEX].status = EXE;
    proc_table[MAIN_INDEX].type = APERIODIC;
    proc_table[MAIN_INDEX].pclass = NRT;
    proc_table[MAIN_INDEX].group = 0;
    proc_table[MAIN_INDEX].dline = PRIO2DLINE(MAIN_PRIORITY);
    proc_table[MAIN_INDEX].priority = PRIO2DLINE(MAIN_PRIORITY);
    proc_table[MAIN_INDEX].period = MAIN_PRIORITY;
    proc_table[MAIN_INDEX].request = 0;
    proc_table[MAIN_INDEX].cap = 0;
    proc_table[MAIN_INDEX].exect = NULL;
    q_extract(MAIN_INDEX,&freedesc);

    /* Fill the dummy() descriptor                      */
    c = VM_context_create(dummy,&dummy_stack[STACK_SIZE],0,task_abort,0);
    proc_table[DUMMY_INDEX].context = c;
    proc_table[DUMMY_INDEX].stack = NULL;
    proc_table[DUMMY_INDEX].body = dummy;
    strcpy(proc_table[DUMMY_INDEX].name,"Dummy");
    proc_table[DUMMY_INDEX].status = READY;
    proc_table[DUMMY_INDEX].type = APERIODIC;
    proc_table[DUMMY_INDEX].pclass = NRT;
    proc_table[DUMMY_INDEX].group = 0;
    proc_table[DUMMY_INDEX].dline = PRIO2DLINE(DUMMY_PRIORITY);
    proc_table[DUMMY_INDEX].priority = PRIO2DLINE(DUMMY_PRIORITY);
    proc_table[DUMMY_INDEX].period = DUMMY_PRIORITY;
    proc_table[DUMMY_INDEX].request = 0;
    proc_table[DUMMY_INDEX].exect = NULL;
    proc_table[DUMMY_INDEX].cap = 0;
    q_extract(DUMMY_INDEX,&freedesc);
    q_insert(DUMMY_INDEX,&ready_nrt);
    load_init();
    #ifdef __NEW_TRACE__
    	/* At system startup, dummy & main are created 	*/
	/* and activated and the main is executed!	*/
	log_record(MAIN_INDEX,ACT_CREATE);
	log_record(DUMMY_INDEX,ACT_CREATE);
	log_record(MAIN_INDEX,ACT_ACTIVATE);
	log_record(MAIN_INDEX,ACT_ACTIVATE);
	log_record(MAIN_INDEX,ACT_DISPATCH);
    #endif
    VM_enable();
   
    /* reset keyboard after exit */
    sys_atexit((void(*)(void))C8042_restore,AFTER_EXIT);
}

/* Used to post an exit function */
int sys_atexit(void (*exit_code)(void),BYTE when)
{
    register i = 0;
    BYTE found = 0;
    SYS_FLAGS f;

    f = VM_save();
    while (!found && i < MAX_EXITFUNC) {
	if (exit_list[i].status == NONE_EXIT) found = 1;
	else i++;
    }
    if (i == MAX_EXITFUNC) {
	proc_table[exec].errno = TOOMUCH_EXITFUNC;
	VM_restore(f);
	return -1;
    }
    exit_list[i].exit_code = exit_code;
    exit_list[i].status = when;
    VM_restore(f);
    return 1;
}

/* Close the system & return to HOST OS. Can be called ONLY into main */
void sys_end(void)
{
    register i;
    /* First process exit-before functions */
    for (i = 0; i < MAX_EXITFUNC; i++)
	if (exit_list[i].status & BEFORE_EXIT) {
	    exit_list[i].status = NONE_EXIT;
	    exit_list[i].exit_code();
	}
    /* Shut down the VM layer */
    /*
       If sys_end() is called by a generic task, then we
       do a clean abort, since we cannot assume anything
       about the stack status...
       Otherwise, it it is main() to call it, then control
       is regained by the HOST OS!! (DOS)
    */
    VM_disable();
    if (exec == MAIN_INDEX) VM_end();
    else VM_abort(GENERIC_ERROR);
    VM_enable();
    /* Now process exit-after functions */
    for (i = 0; i < MAX_EXITFUNC; i++)
	if (exit_list[i].status & AFTER_EXIT) {
	    exit_list[i].status = NONE_EXIT;
	    exit_list[i].exit_code();
	}
}

/* Default exception handlers */
int sys_ignore(int code)
{
    return(1);
}

int sys_abort(int code)
{
    register i;
    /* While the sys_end() returns to main() which can proceed    */
    /* as a standard host OS application, an abort must terminate */
    /* the application immediately                                */
    /* First process exit-before functions                        */
    for (i = 0; i < MAX_EXITFUNC; i++)
	if (exit_list[i].status & BEFORE_EXIT) {
	    exit_list[i].status = NONE_EXIT;
	    exit_list[i].exit_code();
	}
    /* To perform a clean exit we must stand on a safe stack; in fact   */
    /* when the abort exception is detected we do not know what task    */
    /* is executing.                                                    */
    /* The VM_abort() tranfers the control on this stack!               */
    VM_abort(code);
    /* Controls never reach this point                                  */
    /* This line is used to remove the warning...                       */
    return(0);
}

/* The tail code has to be executed on a safe stack, which is set up    */
/* with the VM_abort                                                    */
void __cdecl sys_abort_tail(int code)
{
    register i;
    /* Here the execution resumes from VM_abort() */
    /* Interrupt are already cleared              */
    VM_end();
    VM_enable();
    /* Now process exit-after functions           */
    for (i = 0; i < MAX_EXITFUNC; i++)
	if (exit_list[i].status & AFTER_EXIT) {
	    exit_list[i].status = NONE_EXIT;
	    exit_list[i].exit_code();
	}
    /* If it was a generic error, then just exit */
    /* This is generally intercepted by end user */
    /* so we avoid bothering with a message	 */
    if (code == GENERIC_ERROR) exit(1);
    /* Otherwise display useful informations & exit */
    VM_printf("Abort detected\nCode : %u\n",code);  
    if (code >= CLOCK_OVERRUN && code <= LAST_KERN_ERROR)
       VM_printf("Cause : %s\n",kern_error_msg[code-CLOCK_OVERRUN]);
    else if (code == VM_FAULT) VM_printf("Cause : VM fault\n");
    if (code == WCET_VIOLATION) {
	VM_printf("Task: %s (Pid:%d) Tick: %lu ActTime: %lu Cap: %ld Wcet: %lu\n",
		proc_table[exec].name,exec,sys_tick, act_time,
		-proc_table[exec].cap, proc_table[exec].wcet);
    }
    exit(-1);
}

int errno(void)
{
    return proc_table[exec].errno;
}

const char *sys_err(int err)
{
    if (err >= CLOCK_OVERRUN && err <= LAST_KERN_ERROR)
       return(kern_error_msg[err - CLOCK_OVERRUN]);
    else return("Unknown");
}

void perror(void)
{
    const char *errstr = sys_err(proc_table[exec].errno);
    VM_printf("Error %d --> %s\n", proc_table[exec].errno, errstr);
}

int cbs_guarantee(void)
{
    int i;
    DWORD tot;

    tot = 0;
    for (i = 0; i < MAX_PROC; i++) {
	if ((proc_table[i].status != FREE) &&
		((proc_table[i].pclass == SOFT) || (proc_table[i].pclass == HARD))) {
		    tot += (proc_table[i].wcet * 100) / proc_table[i].period;
		}		    
    }
    if (tot > sys_tick * 100) return -1;
    return 1;
}

/*----------------------------------------------------------------------*/
/* Kernel task management functions                                     */
/*----------------------------------------------------------------------*/

/* Allocate & fill a task descriptor; the task is not explicitely       */
/* activated; you have to use the task_activate to do this              */
/* Just set up the minimum necessary to make thing works well           */
PID task_create(char *name,TASK (*body)(),BYTE pclass,BYTE type,DWORD p,MODEL *m)
{
    PID i = NIL;
    CONTEXT c;
    BYTE *tos;
    WORD ctrl;

    VM_disable();
/*    if (pclass == SOFT && type == PERIODIC) {
	proc_table[exec].errno = UNVALID_CREATE;
	VM_enable();
	return -1;
    }
    else */{
	/* Get a free descriptor */
	i = q_getfirst(&freedesc);
	/* If no one is available abort the system */
	if (i == NIL) {
	    proc_table[exec].errno = NO_AVAIL_TASK;
	    VM_enable();
	    return -1;
	}
	/* Fill in the descriptor */
	strncpy(proc_table[i].name,name,19);
	proc_table[i].name[19] = 0;
	proc_table[i].body = body;
	proc_table[i].stacksize = ((m == NULL || m->stacksize == 0) ? STACK_SIZE : m->stacksize);
	/* Set up control field */
	proc_table[i].control = ctrl = ((m == NULL) ? 0 : m->control);
	proc_table[i].num_act = 0;
	/* Allocate a stack for the task */
	tos = proc_table[i].stack = VM_alloc(proc_table[i].stacksize);
	if (proc_table[i].stack == NULL) {
            q_insert(i,&freedesc);
	    proc_table[exec].errno = NO_AVAIL_STACK_MEM;
	    VM_enable();
	    return -1;
	}
	/* TOS is the TopOfStack! */
	tos += proc_table[i].stacksize;
	/* Task is not active when created */
	proc_table[i].status = SLEEP;
	proc_table[i].type = type;
	proc_table[i].pclass = pclass;
	proc_table[i].next = proc_table[i].prev = NIL;
	/* Some check onto the (pclass,type) of the task  */
	/* Currently HARTIK supports the following models */
	/*
			    HARD    	SOFT    	NRT  (Pclass)
	    (Type)     ----------------------------------------------
	     PERIODIC	Yes (a)		No		Yes (c)
	    APERIODIC	Yes (a)		Yes (b)		Yes (c)
	*/
	/*	(a): HARD tasks are guaranteed on a priori base (reservation)
		 It is necessary to know the wcet to perform the guarantee!

	    (b): SOFT tasks are handled through a server to enhance their
		 responsiveness! Currently the TBS server is used; however
		 it is necessary to know the wcet for this task in order
		 to do not jeopardize the guarantee of the hard tasks.
		 Both soft/hard tasks are placed in the hard_ready queue

	    (c): Non Real-Time tasks (NRT) are handled in background;
		 they are placed in the soft_ready queue and managed
		 using a fixed priority schema
	*/

	/* If the task is APERIODIC/NRT we must translate its priority	*/
	/* into deadline in order to insert them into the queues after      */
	/* any periodic task                                                */

	if (pclass == NRT && type == APERIODIC) p = PRIO2DLINE(p);
	proc_table[i].drel = ((m == NULL || m->drel == 0) ? p : m->drel);
	proc_table[i].dline = p;
	proc_table[i].period = p;
	proc_table[i].group = ((m == NULL) ? 0 : m->group);
	proc_table[i].wcet = ((m == NULL) ? 0 : m->wcet);
	proc_table[i].cap = 0;

 	if ((proc_table[i].wcet == 0) && (pclass != NRT)) {
            if (wcheck || (pclass == SOFT)) {
            VM_free(proc_table[i].stack,proc_table[i].stacksize);
	    proc_table[exec].errno = CANNOT_GUARANTEE;
	    proc_table[i].status = FREE;
	    q_insert(i,&freedesc);
	    VM_enable();
	    return -1;
            }
	}

	if (pclass == HARD) {
//	    proc_table[i].wcet = proc_table[i].wcet + sys_tick;
	}
	/* For CBS tasks, we have to fill some others fields...	*/
	if (pclass == SOFT) {
	    if (proc_table[i].wcet < (sys_tick / MIN_Q)) {
		proc_table[i].wcet = (sys_tick / MIN_Q);
	    }
	    proc_table[i].q = proc_table[i].wcet;
	}

	/* exec time extimation... */
	if (m == NULL) proc_table[i].exect = NULL;
    	else proc_table[i].exect = m->exect;

	#ifdef __SRP__
	    /* Setup the resource table for the task */
	    if (m != NULL)
	       memcpy(srp_block[i].section,m->res,sizeof(DWORD)*MAX_RESOURCE);
	    else memset(srp_block[i].section,0,sizeof(DWORD)*MAX_RESOURCE);
	    /* Test if we need a guarantee on this task */
	    /* Warning! Need to use floats!             */
#ifdef __TBS__
	    if (pclass == HARD || pclass == SOFT) {
#else
	    if (pclass == HARD) {
#endif
		if (srp_guarantee(i) == -1) {
                    VM_free(proc_table[i].stack,proc_table[i].stacksize);
		    proc_table[exec].errno = CANNOT_GUARANTEE;
		    proc_table[i].status = FREE;
		    q_insert(i,&freedesc);
		    VM_enable();
		    return -1;
		}
	    }
	#else
	    if (pclass == HARD || pclass == SOFT) {
		if (cbs_guarantee() == -1) {
                    VM_free(proc_table[i].stack,proc_table[i].stacksize);
		    proc_table[exec].errno = CANNOT_GUARANTEE;
		    proc_table[i].status = FREE;
		    q_insert(i,&freedesc);
		    VM_enable();
		    return -1;
		}
	    }
	#endif

	/* Set up the initial context */
	c = VM_context_create(body,tos,((m == NULL) ? 0 : m->arg),task_abort,ctrl);
	proc_table[i].context = c;
	/* Zero all other stuff descriptor parameters */
	#ifdef __HOOK__
	    act_exc(CREATE_HOOK);
	#endif
	#ifdef __TRACE__
	    if ( (m == NULL) || (m->control & TRACE_TASK) )
		    log_record(i,ACT_CREATE);
	    else log_record(i, ACT_CREATE_NO_TRACE);
	#endif
	VM_enable();
    }
    return(i);
}

static void task_makefree(PID i)
{
    /* Free the used context space */
    VM_context_delete(proc_table[i].context);
    #ifdef __TRACE__
	log_record(i,ACT_KILL);
    #endif    
    /* Free the used stack area */
    /* Warning! we could be executing on the stack we are */
    /* going to destroy; this should cause no problem as  */
    /* the task_kill() cannot be interrupted & the memory */
    /* released can only be reassigned when we exit the   */
    /* primitive!                                         */
    VM_free(proc_table[i].stack,proc_table[i].stacksize);
    /* Check to see in what queue is the task */
    switch (proc_table[i].status) {
	case WAIT :     q_extract(i,&blocked);
			break;
	case IDLE :     q_extract(i,&idle);
			break;
	case READY :    if (proc_table[i].pclass == HARD || proc_table[i].pclass == SOFT)
				q_extract(i,&ready_rt);
			else q_extract(i,&ready_nrt);
			break;
	case EXE :      dispatch();
			break;
    }

   switch(proc_table[i].pclass) {
      case HARD:
      case SOFT:
	if (proc_table[i].dline>sys_clock) {
	    proc_table[i].priority = proc_table[i].dline;
	    proc_table[i].status = ZOMBIE;
	    q_insert(i,&zombie);
	} else
	  _endtask(i);
	break;
      case NRT:
	_endtask(i);
	break;
    }    

}

/* This primitive autokills the excuting task; it is used to avoid      */
/* that returning from a task cause a jmp to an unpredictable location  */
void task_abort(void)
{
    VM_disable();
    if (exec == DUMMY_INDEX || exec == MAIN_INDEX) act_exc(UNVALID_KILL);
    task_makefree(exec);
    #ifdef __HOOK__ 
	act_exc(ABORT_HOOK);
    #endif    
    VM_context_load(proc_table[exec].context);
}

/* This primitive kills the i-th task, free the descriptor, and the     */
/* memory for the stack                                                 */
int task_kill(PID i)
{
    BYTE autokill;

    VM_disable();
    /* Detect main/dummy killing */
    if (i == DUMMY_INDEX || i == MAIN_INDEX || proc_table[exec].status == FREE) {
	proc_table[exec].errno = UNVALID_KILL;
	VM_enable();
	return -1;
    }
    /* Detect suicide */
    autokill = (i == exec);
    task_makefree(i);
    #ifdef __HOOK__ 
	act_exc(KILL_HOOK);
    #endif    
    /* If we were killing ourselves we must switch context */
    /* without saving the actual one                       */
    if (autokill) VM_context_load(proc_table[exec].context);
    else VM_enable();
    return 1;
}

/* Suspend the executing task till the next period                      */
void task_endcycle(void)
{
    CONTEXT c;
    /* The executing task will be now placed into idle state  */
    /* if it is periodic or in sleep state if it is aperiodic */
    /* If the task is an hard-periodic task & we are calling  */
    /* the task_endcycle, then it cannot be over the deadline */
    /* because the timer should detect this; anyway it could  */
    /* be finishing just on the deadline! In this case, we    */
    /* must place it into the ready queue                     */

    c = VM_context_save();
    proc_table[exec].context = c;

    /* First, we have to manage the capacity of the running task...	*/
    /* 		----By Luca	:-(					*/

    CBS_PROLOGUE();

    /* exec time est.*/
    if (proc_table[exec].exect != NULL) {
	*(proc_table[exec].exect) = (DWORD)(-proc_table[exec].cap);
	proc_table[exec].exect = &(proc_table[exec].exect[1]);
    }

    if (proc_table[exec].pclass != SOFT) proc_table[exec].cap = 0;

    switch (proc_table[exec].type) {
	case APERIODIC : if (proc_table[exec].pclass == SOFT &&
			proc_table[exec].num_act > 0) {
		/* 4 CBS tasks, the dline-priority and capacity are managed at the
	   	moment of scheduling a task... */
	    	proc_table[exec].num_act--;
		proc_table[exec].request = act_time;
		proc_table[exec].status = READY;
		q_insert(exec,&ready_rt);
	    } 
	    else proc_table[exec].status = SLEEP;
	    break;
	case PERIODIC :
	     if ((sys_clock < proc_table[exec].dline) && (proc_table[exec].pclass != SOFT)) {
		 proc_table[exec].status = IDLE;
		 proc_table[exec].request += proc_table[exec].period;
		 proc_table[exec].dline = proc_table[exec].request + proc_table[exec].drel;
		 proc_table[exec].priority = proc_table[exec].request;
		 q_insert(exec,&idle);
	     } else if (sys_clock < proc_table[exec].request + proc_table[exec].period) {
		 proc_table[exec].status = IDLE;
		 proc_table[exec].request += proc_table[exec].period;
		 proc_table[exec].priority = proc_table[exec].request;
		 q_insert(exec,&idle);
	     }
	     else {
		 if (proc_table[exec].pclass == HARD) {
		     proc_table[exec].request += proc_table[exec].period;
		     proc_table[exec].dline = proc_table[exec].request + proc_table[exec].drel;
		     proc_table[exec].priority = proc_table[exec].dline;
		     proc_table[exec].status = READY;
		     q_insert(exec,&ready_rt);
		 }
	/* 4 CBS tasks, the dline-priority and capacity are managed at the
	   moment of scheduling a task... */
		 else if (proc_table[exec].pclass == SOFT) {
		     proc_table[exec].request += proc_table[exec].period;
		     proc_table[exec].status = READY;
		     q_insert(exec,&ready_rt);
		 }
		 /* It cannot happen that a task is SOFT/PERIODIC!  */
		 /* So this else statement refers only to NRT tasks */
		 else {
		     while (proc_table[exec].request <= sys_clock)
			proc_table[exec].request += proc_table[exec].period;
		     proc_table[exec].dline = proc_table[exec].request + proc_table[exec].drel;
		     proc_table[exec].priority = proc_table[exec].request;
		     proc_table[exec].status = IDLE;
		     q_insert(exec,&idle);
		}
	     } 
	     break;
    }
    #ifdef __TRACE__
	log_record(exec,ACT_ENDCYCLE);
    #endif
    dispatch();
    #ifdef __HOOK__ 
	act_exc(ENDCYCLE_HOOK);
    #endif
    VM_context_change(proc_table[exec].context);
}

/* Puts the p-th task in the ready queue */
void task_makeready(PID p)
{
    /* Manage SRP activations */
    if (proc_table[p].status == LOBBY) {
	proc_table[p].status = LOBBY_READY;
	return;
    }
    /* Test if we are trying to activate a non sleeping task    */
    /* Ignore this; the task is already active                  */
    if (proc_table[p].status != SLEEP) {
	if (proc_table[p].pclass != SOFT) return;
	proc_table[p].num_act++;
	return;
    }
    /* Test for hard periodic task or CBS task */
    proc_table[p].request = sys_clock;
    if (proc_table[p].pclass == HARD)
	proc_table[p].dline = sys_clock + proc_table[p].drel;
    else if (proc_table[p].pclass == SOFT) {
	/* When a new CBS task arrives to the system, we must check the
	   residual CPU band of the CBS... */
	if (proc_table[p].dline * sys_tick < act_time +
		(proc_table[p].cap * proc_table[p].period * sys_tick) / proc_table[p].q){
		    proc_table[p].cap = proc_table[p].q;
		    proc_table[p].dline = proc_table[p].request + proc_table[p].period;
		    proc_table[p].priority = proc_table[p].request + proc_table[p].period;
	} 
    } else if (proc_table[p].type == PERIODIC)
	   proc_table[p].dline = sys_clock + proc_table[p].drel;

    /* Insert task in the correct queue */
    proc_table[p].priority = proc_table[p].dline;
    proc_table[p].status = READY;
    if (proc_table[p].pclass == HARD || proc_table[p].pclass == SOFT)
       q_insert(p,&ready_rt);
    else q_insert(p,&ready_nrt);
}

/* Activates a single task */
int task_activate(PID p)
{
    /* Ma un po' di controlli??? */
    if (p > MAX_PROC) {
	proc_table[exec].errno = UNVALID_TASK_ID; /* trovare codice errore */
	return -1;
    }
    if (proc_table[p].status == FREE) {
	proc_table[exec].errno = UNVALID_TASK_ID; /* trovare codice errore */
	return -1;
    }

    /* Begin activate */
    proc_table[exec].context = VM_context_save();

    /* First, we have to manage the capacity of the running task...	*/
    /* 		----By Luca	:-(					*/
    CBS_PROLOGUE();

    task_makeready(p);
    #ifdef __TRACE__
	log_record(p,ACT_ACTIVATE);
    #endif
    /* Preempt if necessary */
    scheduler();
    #ifdef __HOOK__ 
	act_exc(ACTIVATE_HOOK);
    #endif    
    VM_context_change(proc_table[exec].context);
    return 1;
}

void task_nopreempt(void)
{
    VM_disable();
    proc_table[exec].control |= NO_PREEMPT;
    VM_enable();
}

void task_preempt(void)
{
    proc_table[exec].context = VM_context_save();
    /* First, we have to manage the capacity of the running task...	*/
    /* 		----By Luca	:-(					*/
    CBS_PROLOGUE();

    proc_table[exec].control &= ~NO_PREEMPT;
    scheduler();
    VM_context_change(proc_table[exec].context); 
}

/* Puts the actual executing task in the sleep state */
/* Perhaps we should add a check for NRT tasks only! */
void task_sleep(void)
{
    proc_table[exec].context = VM_context_save();
    proc_table[exec].status = SLEEP;
    #ifdef __TRACE__
	log_record(exec,ACT_SLEEP);
    #endif
    dispatch();
    #ifdef __HOOK__ 
	act_exc(SLEEP_HOOK);
    #endif    
    VM_context_change(proc_table[exec].context);
}

/* Suspend the task for s time unit                                     */
void task_delay(DWORD tickdelay)
{
    proc_table[exec].context = VM_context_save();
    proc_table[exec].status = IDLE;
    proc_table[exec].priority = sys_clock + tickdelay;
    q_insert(exec,&idle);
    #ifdef __TRACE__
	log_record(exec,ACT_DELAY);
    #endif
    dispatch();
    #ifdef __HOOK__
	act_exc(DELAY_HOOK);
    #endif    
    VM_context_change(proc_table[exec].context);
}

int group_activate(WORD g)
{
    register PID i;
    
    if (g == 0) {
	proc_table[exec].errno = UNVALID_GROUP;
	return -1;
    }
    proc_table[exec].context = VM_context_save();

    CBS_PROLOGUE();

    for (i=0 ; i <  MAX_PROC; i++)
        if (proc_table[i].group == g) {
	    task_makeready(i);
	    #ifdef __TRACE__
		log_record(i,ACT_ACTIVATE);
	    #endif
	}
    scheduler();
    #ifdef __HOOK__ 
	act_exc(GROUP_ACT_HOOK);
    #endif
    VM_context_change(proc_table[exec].context);
    return 1;
}
		
void group_kill(WORD g)
{
    BYTE autokill;
    PID i;

    if (g == 0) exc_raise(UNVALID_GROUP);
    VM_disable();
        
    /* Detect suicide */
    autokill = (g == proc_table[exec].group);
    for (i=0 ; i <  MAX_PROC; i++)
	if (proc_table[i].status != FREE && proc_table[i].group == g)
	    task_makefree(i);
    #ifdef __HOOK__
       act_exc(GROUP_KIL_HOOK);
    #endif
    /* If we were killing ourselves we must switch context */
    /* without saving the actual one                       */
    if (autokill) VM_context_load(proc_table[exec].context);
    else VM_enable();
}

/*----------------------------------------------------------------------*/
/* Kernel utility/debug functions                                       */
/*----------------------------------------------------------------------*/

void print_queue(QUEUE q)
{
    while (q != NIL) {
	VM_printf("Pid: %d\t Name: %s\t DL: %lu\n",q,proc_table[q].name,proc_table[q].dline);
	q = proc_table[q].next;
    }
}

void sys_status(DWORD cw)
{
    register q;
    SYS_FLAGS f;

    f = VM_save();
    if (cw & CLOCK_STATUS) VM_printf("Sys clock : %lu\n",sys_clock);
    if (cw & EXEC_STATUS) {
	VM_printf("< EXEC > \n");
	VM_printf("Pid : %d \tName : %s\treq : %lu\tdl : %lu\n",exec,
					proc_table[exec].name,
					proc_table[exec].request,
					proc_table[exec].dline);
    }
    if (cw & READY_RT_STATUS) {
	VM_printf("< READY RT >\n");
	print_queue(ready_rt);
    }
    if (cw & READY_NRT_STATUS) {
	VM_printf("< READY NRT >\n");
	print_queue(ready_nrt);
    }
    if (cw & BLOCKED_STATUS) {
	VM_printf("< BLOCKED QUEUE >\n");
	print_queue(blocked);
    }
    if (cw & IDLE_STATUS) {
	VM_printf("< IDLE >\n");
	print_queue(idle);
    }
    if (cw & SLEEP_STATUS) {
	VM_printf("< SLEEP >\n");
	for (q = 0; q < MAX_PROC; q++) {
	    if (proc_table[q].status == SLEEP) {
		VM_printf("Pid: %d\tName: %s\tdl: %lu\n",q,
					proc_table[q].name,
					proc_table[q].dline);
	    }
	}
    }
    if (cw & MEM_STATUS) VM_mem_dump();
    #ifdef __PORT__
	if (cw & PORT_STATUS) print_port();
    #endif
    VM_restore(f);
}

/* Read the system clock; the value is converted into usec */

TIME sys_time(void)
{
    DWORD r1;
    SYS_FLAGS f;
    
    f = VM_save();
    r1 = VM_timer_read();
    VM_restore(f);
    return(r1);
}

DWORD sys_ticks(void)
{
    DWORD r1;
    SYS_FLAGS f;
    
    f = VM_save();
    r1 = sys_clock;
    VM_restore(f);
    return(r1);
}

DWORD sys_gettick(void)
{
    DWORD r1;
    SYS_FLAGS f;

    f = VM_save();
    r1 = sys_tick;
    VM_restore(f);
    return r1;
}

/* Well this function could be not necessary as the kernel structures   */
/* are visible from the application.					*/
/* Anyway we encourage the use of this function, hoping that in the	*/
/* next future, the kernel will become a separate system		*/

PID task_exec(void)
{
    return(exec);
}

PID task_pid(char *name)
{
    PID i,res = NIL;
    SYS_FLAGS f;
    
    f = VM_save();
    for (i = 0; i < MAX_PROC; i++)
	 if ((proc_table[i].status != FREE) &&
	     (strcmp(name, proc_table[i].name) == 0)) res = i;
    VM_restore(f);

    return(res);
}

char *task_name(PID p)
{
    return(proc_table[p].name);
}

DWORD task_dline(PID p)
{
    return(proc_table[p].dline);
}

BYTE task_status(PID p)
{
    return(proc_table[p].status);
}

/* Last but not least, here is the function used to get the difference  */
/* between the deadline assigned by the CBS to a task (scheduling	*/
/* deadline) and the soft deadline of the task				*/
/* This value is usefull to make an application adapt to the reserved	*/
/* bandwith (simple, no?)						*/
/*						By Luca :-(		*/

DWORD task_howlate(PID p)
{
    return(proc_table[p].dline - (proc_table[p].request + proc_table[p].period));
/*    return proc_table[p].request;*/
}

#ifdef __KERN_DEBUG__
    #include "kernstub.c"
#endif
