/* Project:     HARTIK 3.1.2                                    */
/* Description: Hard Real TIme Kernel for 8086 compatible       */
/* Author:      Gerardo Lamastra                                */
/*              Renato Rozzi                                    */
/* Date:        01/07/98                                        */

/* File:        Fix_log.C                                       */
/* Revision:    1.0                                             */

#include <string.h>
#include <mem.h>
#include <stdlib.h>
#include <xdos.h>

#include "vm.h"
#include "kern.h"
#include "log.h"

#define __DEBUG__

/* Just to have a look at log-overhead 		*/
static DWORD max_record_length = 0,max_reset_length=0,max_fix_length=0;
static DWORD record_length, reset_length, fix_length;

/* The LOG_INFO parameters 			*/
/* fix_dim = nr of entries of fix table 	*/
/* loop_dim = nr of entries of loop table	*/
static DWORD fix_dim,loop_dim;

/* Start Trace Mode {LOOP,FIX} & trace-file name */
static BYTE start_mode;
static char *name;


/* The variables used to hold the log 	*/ 
/* Pointer to fix_table 		*/
static LOG_DES* fix_table;
/* Pointer to loop_table 		*/
static LOG_DES* loop_table;

/* To store creation time */
static DWORD creation_times[MAX_PROC];

/* Used to save last events extracted from loop_table 	*/
static BYTE mem_table[MAX_PROC];
/* Used to save status at reset time			*/
static BYTE reset_states[MAX_PROC];
/* Current status of tasks alive */
static BYTE cur_states[MAX_PROC];
/* Count setted bytes in cur_states */
static BYTE task_cnt;
/*flags to know if we must store info about tasks in fix table*/
static BYTE need_info[MAX_PROC];
/* count setted bytes in need_info */
static BYTE info_cnt;
/* info_cnt <= task_cnt ALWAYS */
/* if need_info[i] --> cur_states[i] */
/* Beginning of loop window */
static DWORD loop_start;
static DWORD start_time;
static DWORD stop_time = MAX_DWORD;
static DWORD reset_time = MAX_DWORD;
static BYTE reset_start_mode;
static DWORD fix_ptr, fix_cnt;
/* To count fix_entries I will need in future */
static DWORD fut_need;
static DWORD loop_ptr,loop_cnt,loop_base;
/* Fix windows counter */
static BYTE fw_cnt;
/* Next fix window we'll meet extracting events from loop_table*/
static BYTE nfw;
static BYTE mode = 0;
/* Set if we can't insert any other record in fix_table */
static BYTE no_fix;

static LOG_DES grecord;
static LOG_DATA info;

#define SIZE sizeof(LOG_DES)
#define DWS sizeof(DWORD)
/* UPDATE_DIM * SIZE is the smaller number >= (MAX_PROC + DWS) */
/* and multiple of SIZE */
#define UPDATE_DIM 10

/* Used to mask tracing of tasks */
static BYTE mask[MAX_PROC];

#define FIX_DIM_MIN  500
#define LOOP_DIM_MIN 500

static char *log_error_msg[] = {
	"Out of memory allocating log table",
	"No-consistent parameters to log_init",
	"Cannot open the log file",
	"Log Dump error!",
	"Cannot write to the log file"
};


static int log_exception(int code)
{
	VM_printf("\nLog exception\n");
	VM_printf("Code [%d]\nCause : %s\n",code,log_error_msg[code-LOG_OUT_OF_MEM]);
	/* Log default action should be an ignore! */
	return(0);
}

/* This function initializes the variables used to hold the log    */
/* It should be callled from sys_init rather than user application */

void log_init(LOG_INFO *l)
{
    register i;
	
    if (mode) return;        

    VM_disable();
    for (i = LOG_OUT_OF_MEM; i <= LOG_CANNOT_WRITE; i++)
        exc_set(i,log_exception);
    fix_dim = l->fix_events + l->fix_windows * (UPDATE_DIM + 2);
    loop_dim = l->loop_events;
    /* Let the application choose how to start */
    mode = l->start_mode;
    /* Always need a loop */
    loop_dim = max(loop_dim, LOOP_DIM_MIN);
    loop_cnt = 0; loop_base = 0; loop_ptr = 0;
    fw_cnt = 0; 
    fix_cnt = 0;                
    fut_need = 0;
    nfw = 0;
    no_fix = 0;
    task_cnt = 0;
    info_cnt = 0;
    switch(mode) {
	case LOG_FIX: fix_dim = max(fix_dim, FIX_DIM_MIN);
		      break;
	case LOG_LOOP: if (fix_dim == 0) no_fix = 1;
		       else fix_dim = max(fix_dim, FIX_DIM_MIN);
		       break;
	default : exc_raise(LOG_INIT_ERROR);
    }
    if (fix_dim > 0) {
	fix_table = (LOG_DES*) VM_alloc(fix_dim * SIZE);
	if (fix_table == NULL)  exc_raise(LOG_OUT_OF_MEM);
	fix_ptr = fix_dim - 1;  
	fut_need += 1;  /* You could need a act_no_fix */
    }
    loop_table = (LOG_DES*) VM_alloc(loop_dim * SIZE);
    if (loop_table == NULL)  exc_raise(LOG_OUT_OF_MEM);
    name = l->name;
    start_mode = l->start_mode;
    memset(mem_table,0,MAX_PROC);
    memset(cur_states,0,MAX_PROC);
    memset(need_info,0,MAX_PROC);
    start_time = VM_timer_read();
    if (mode == LOG_FIX) {
	 /* Don't count the initial fix window */
	 fut_need += 9;   /* I will need info MAIN, DUMMY and END_WINDOW */
	 need_info[MAIN_INDEX] = 1;
	 fix_table[fix_ptr].t = start_time;    
	 fix_table[fix_ptr].pid = MAIN_INDEX;    
	 fix_table[fix_ptr].act = ACT_CREATE;    
	 fix_table[fix_ptr].st = EXE;    
	 fix_ptr--;
	 
	 need_info[DUMMY_INDEX] = 1;
	 fix_table[fix_ptr].t = start_time;    
	 fix_table[fix_ptr].pid = DUMMY_INDEX;    
	 fix_table[fix_ptr].act = ACT_CREATE;    
	 fix_table[fix_ptr].st = READY;    
	 fix_ptr--;
	 
	 fix_cnt += 2;
	 info_cnt += 2;
	 loop_start = MAX_DWORD;    
    }    
    else {
	loop_table[loop_ptr].t = start_time;    
	loop_table[loop_ptr].pid = MAIN_INDEX;
	loop_table[loop_ptr].act = ACT_CREATE;    
	loop_table[loop_ptr].st = EXE;    
	loop_ptr++;
	
	loop_table[loop_ptr].t = start_time;    
	loop_table[loop_ptr].pid = DUMMY_INDEX;    
	loop_table[loop_ptr].act = ACT_CREATE;    
	loop_table[loop_ptr].st = READY;    
	loop_ptr++;
	
	loop_cnt += 2;
	loop_start = start_time;
    }
    cur_states[MAIN_INDEX] = EXE;
    cur_states[DUMMY_INDEX] = READY;        
    creation_times[MAIN_INDEX] = creation_times[DUMMY_INDEX] = start_time;
    mask[MAIN_INDEX] = MAIN_INDEX;    /* Don't mask MAIN and DUMMY */
    mask[DUMMY_INDEX] = DUMMY_INDEX;
    task_cnt += 2;

    sys_atexit(log_dump,AFTER_EXIT);
    VM_enable();
};    

/* Reset the fix and the loop tables */
void log_reset()
{
    if (mode & LOG_STOP) return;
    
    VM_disable();
    #ifdef __DEBUG__
	reset_length = VM_timer_read();
    #endif
    
    reset_time = VM_timer_read();
    loop_cnt = 0; loop_base = 0; loop_ptr = 0;  /* reset loop_table */
    fw_cnt = 0; fix_cnt = 0; fut_need = 0;                
    nfw = 0;
    if (fix_dim > 0) { /* Reset fix_table if any*/
	fix_ptr = fix_dim - 1;  
	fut_need = 1; /* you could need a act_no_fix */
	no_fix = 0;
    }
    else no_fix = 1;
    /* Reset start mode */
    mode = start_mode;
    if (mode == LOG_FIX) {	
	/* check if I can start with a fix 				*/
	/* I will need END, FIX_FULL, INFO of task alive at reset time 	*/
	if (fix_dim > (2 + task_cnt * 4) ) {	    
	    memcpy(reset_states, cur_states, MAX_PROC);
	    memcpy(need_info, cur_states, MAX_PROC);
	    memset(mem_table, 0, MAX_PROC);
	    info_cnt = task_cnt;
	    loop_start = MAX_DWORD;
	    fut_need += (info_cnt * 4) + 1;
	} else mode = LOG_LOOP;
    }
    if (mode == LOG_LOOP) {
	memcpy(mem_table, cur_states, MAX_PROC);
	memset(need_info, 0, MAX_PROC);
	info_cnt = 0;
	loop_start = reset_time;
    }
    reset_start_mode = mode;
    #ifdef __DEBUG__
	reset_length = VM_timer_read() - reset_length;
	if (reset_length > max_reset_length)
	   max_reset_length = reset_length;
    #endif   
    VM_enable();
}

/* Insert the record into the loop table */
void loop_insert(LOG_DES* l)
{        
    BYTE delta;
    LOG_DES r;        
	
    /* If the queue is full, drop the first packet 	*
     * and save new state in mem_table			*/
    if (loop_cnt == loop_dim) {
       r = loop_table[loop_base]; 
       delta = 1;
       switch (r.act) {	   
	   case ACT_KILL:
	   	mem_table[r.pid] = 0;
	      	if (loop_table[(loop_base+1)%loop_dim].act == ACT_INFO)
	      	delta += 4;
	      	break;
	   default:
	      	mem_table[r.pid] = r.st;
	      	break;
       }
       loop_base = (loop_base + delta) % loop_dim;
       loop_cnt -= delta;
       loop_start = loop_table[loop_base].t;
       
       /* If we have met the next fix window.....*/
       if ( (nfw) && 
	    (loop_table[loop_base].t > fix_table[(nfw-1)*UPDATE_DIM].t)) {
	        memcpy(&fix_table[(nfw-1)*UPDATE_DIM].pid,mem_table,MAX_PROC);
		fix_cnt += UPDATE_DIM;
		fut_need -= UPDATE_DIM;
		memset(mem_table,0,MAX_PROC);
		if (nfw < fw_cnt) nfw++;
		else nfw = 0;
	    }
    }
    loop_cnt++;
    memcpy(&loop_table[loop_ptr],l,SIZE);            
    loop_ptr = (loop_ptr + 1) % loop_dim;
}

/* Insert the record into the fix table */
void fix_insert(LOG_DES* l)
{         
    /* Fix_insert always possible */
    memcpy(&fix_table[fix_ptr],l,SIZE);
    fix_ptr--;
    fix_cnt++;
}   

/* Insert the action FIX_FULL into the fix table */
void act_no_fix()
{
    grecord.t = VM_timer_read();
    grecord.act = ACT_FIX_FULL;
    memcpy(&fix_table[fix_ptr],&grecord, SIZE);
    fix_ptr--;
    fix_cnt++;
    fut_need--;
}

/* Stop recording events in log tables */
void log_stop()
{
    SYS_FLAGS f;
    f = VM_save();
    stop_time = VM_timer_read();
    mode |= LOG_STOP;
    VM_restore(f);
}

/* Switch to loop mode --> next records will be */
/* inserted into loop_table and could be lost   */
void log_loop(void)
{
    SYS_FLAGS f;
    
    if ( (mode & LOG_STOP) || (mode & LOG_LOOP)) return;
    f = VM_save();
    grecord.t = VM_timer_read();
    if (loop_cnt == 0) loop_start = grecord.t;
    grecord.act = ACT_END_WINDOW;  /*end of previous fix window */
    fix_insert(&grecord);
    fut_need--;
    mode = LOG_LOOP;
    VM_restore(f);
}

/* Switch to fix mode --> next records will 	*/
/* be inserted into fix_table and won't be lost */
void log_fix(void)
{    
    SYS_FLAGS f;
    BYTE delta;
    
    if ( (mode & LOG_STOP) || (mode & LOG_FIX) || no_fix) return;   

    f = VM_save();
    #ifdef __DEBUG__
	fix_length = VM_timer_read();
    #endif
    /* I will need END, update_info, info of tasks absent in fix_table */
    delta = 1 + UPDATE_DIM + (task_cnt - info_cnt) * 4;
    if ((fix_cnt + fut_need + delta + 1) >= fix_dim ) {
	/* 1 for START*/
	no_fix = 1;
	act_no_fix();
    }
    else {
	mode = LOG_FIX;
	fw_cnt++;
	if (!nfw) nfw = fw_cnt;
	grecord.t = VM_timer_read();
	memcpy(&fix_table[(fw_cnt - 1) * UPDATE_DIM].t,&grecord.t,DWS);
	grecord.act = ACT_START_WINDOW;
	fix_insert(&grecord);
	fut_need += delta;
	memcpy(need_info, cur_states, MAX_PROC);
	info_cnt = task_cnt;
    }
    #ifdef __DEBUG__
	fix_length = VM_timer_read() - fix_length;
	if (fix_length > max_fix_length) max_fix_length = fix_length;
    #endif       
    VM_restore(f);
}

/* This is usally called from a primitive, so it is not protected       */
/* with VM_disable/VM_enable! If the application can use it, protection */
/* should be enforced. This is a problem similar to VM_alloc/VM_free    */
/* management... A single solution would certainly be better...         */

/* In case of FASTON (FASTOFF) act = FASTON_FLAG(FASTOFF_FLAG) | IRQ#      */
/* In log tables I store a classic HARTIK status (e.g. EXE, SLEEP) except  */
/* in case of FAST (status = EXE), just to make faster the extraction from */
/* loop_table in case it is full (in mem_table I have just 1 byte to store */
/* there's a fast and it's number                                          */
void log_record(PID p,BYTE act)
{
    register int i;
    SYS_FLAGS f;
    BYTE delta;
    BYTE go_info = 1;
    LOG_DES record;
    BYTE pid=0;  

    f = VM_save();
   
    #ifdef __DEBUG__
	record_length = VM_timer_read();
    #endif

    if (! (mode & LOG_STOP)) {
	/* Must insert record somewhere */
	record.t = VM_timer_read();
	switch (act) {
	    case ACT_CREATE_NO_TRACE:
	    	record.act = ACT_CREATE;
		mask[p] = record.pid = DUMMY_INDEX;  /* Mask the task */
		record.st = proc_table[p].status;
		cur_states[DUMMY_INDEX] = record.st;
		break;
	    case ACT_CREATE: 
		task_cnt++;
		creation_times[p] = record.t;
		mask[p] = p;
	    default:
		pid = mask[p];
		record.pid = pid;
		record.act = act;
		if (act == ACT_KILL) {
		    record.st = FREE;
		    /* It should be so (check makefree in kern.c) */
		    if (pid == p) {
			cur_states[p] = 0;
			task_cnt--;
		    } else go_info = 0;
		}
		else if (act & FAST_FLAG) record.st = cur_states[pid] = act;
		else record.st = cur_states[pid] = proc_table[p].status; 
		break;
	}
		
	delta = (need_info[pid] == 0) ? 4 : 0;
	if ( (mode & LOG_FIX) && 
	     ((fix_cnt + fut_need + delta) >= fix_dim) ) {
	    no_fix = 1;
	    act_no_fix();
	    log_loop();
	}
	if (mode & LOG_FIX) {
	    fix_insert(&record);
	    if (delta) {
		fut_need += delta;
		need_info[pid] = 1;
		info_cnt++;
	    }
	} else loop_insert(&record);
    }
    else go_info = cur_states[p];  
    if (( act == ACT_KILL) && go_info)  {	
	/* Must save other info...*/
	if (mode & LOG_STOP ) record.t = stop_time;
	record.pid = p;
	record.act = ACT_INFO;
	record.st = FREE;
	memcpy(info.name, proc_table[p].name, NAME_LEN);
	info.creation = creation_times[p];
	info.deadline = proc_table[p].drel * sys_tick;
	info.type_class = (proc_table[p].type << 4) | proc_table[p].pclass;
	info.period = proc_table[p].period * sys_tick;
	if (need_info[p]) {
	    memcpy(&fix_table[fix_ptr], &record, SIZE);
	    memcpy(&fix_table[fix_ptr - 3], &info, sizeof(LOG_DATA));
	    fix_ptr -= 4;
	    fix_cnt += 4;
	    fut_need -= 4;
	    info_cnt--;
	}
	else {
	    loop_insert(&record);
	    for (i=0;i<3;i++)
	        loop_insert((LOG_DES*)((BYTE*)&info + i * SIZE));
	}
	need_info[p] = 0;
	cur_states[p] = 0;
    }
    #ifdef __DEBUG__
	record_length = VM_timer_read() - record_length;
	if (record_length > max_record_length)
	    max_record_length = record_length;
    #endif
    VM_restore(f);
}
	
/* This is the dump system... It will probably change in the future      */
/* Now they are mapped onto the DOS_open, DOS_close, DOS_write functions */
/* which are available in the basic standard library...                  */
/* This is dirt, but we are waiting for filesystem support (Sorry!)      */
static DOS_FILE *log_file = NULL;

unsigned lopen(char *name)
{
    log_file = DOS_fopen(name,"wb");
    if (log_file == NULL) return(0);
    else return(1);
}

void lclose(void)
{
    DOS_fclose(log_file);
}

unsigned lwrite(void *data,DWORD size)
{
    if (DOS_fwrite(data,size,1,log_file) != size) return(0);
    else return(1);
}

/* Log save use the abstact I/O interface to save a single entry */
/* of the log_table. It is not necessary to make it global.	 */

int log_save(DWORD pos, int dir)
{
    BYTE i;
    LOG_DES* log_table;

    if (dir > 0) log_table = loop_table;
    else log_table = fix_table;

    if (log_table[pos].act != ACT_INFO) {
	if (!lwrite(&log_table[pos],SIZE)) exc_raise(LOG_CANNOT_WRITE);
	return 1;
    }
    /* Otherwise we must store on disk 4 records */
    if (dir > 0) for (i=0;i<4;i++) {
	    if (!lwrite(&log_table[pos],SIZE) ) exc_raise(LOG_CANNOT_WRITE);
	    pos = (pos + 1) % loop_dim;
    }
    else {
	if (!lwrite(&log_table[pos], SIZE)) exc_raise(LOG_CANNOT_WRITE);
	if (!lwrite(&log_table[pos - 3], sizeof(LOG_DATA)) )
	    exc_raise(LOG_CANNOT_WRITE);
    }
    return 4;
}

/* This is the dump_function... it is usually called after the system	*/
/* has been closed. However, when the filesystem will be available, it	*/
/* could be called at any time! 					*/
void log_dump(void)
{       	
    register BYTE active = 0;        
    DWORD fixScan = fix_dim - 1;
    DWORD loopScan = loop_base;
    DWORD loopSaveCount = loop_cnt;
    DWORD fixSaveCount = fix_cnt;
    DWORD delta, last_clock = MAX_DWORD;
    BYTE cfw = 0; /* Current fix window */
    BYTE i;
    BYTE final_window = 0; /* To know if we are in the final window */
    BYTE *ptr;
       
    /* Please, don't worry about how I handle actions like ACT_START_WINDOW */
    /* ACT_END_WINDOW. They are just used to let the tracer know when a     */
    /* window starts or finishes.                                           */
        
    #ifdef __DEBUG__
    	VM_printf("Start dump info\n");
	VM_printf("Fix_dim %d\tloop_dim %d\n",fix_dim,loop_dim);
	VM_printf("Fw_cnt %d\tno_fix %d\n",fw_cnt,no_fix);
	VM_printf("Fix_cnt %d\tLoop_cnt %d\tLoop_base %d\n",
		  fix_cnt,loop_cnt,loop_base);
	VM_printf("Loop start = %d\n", loop_start);
	VM_printf("Stop time  = %d\n", stop_time);
	VM_printf("Reset time  = %d\n", reset_time);
	VM_printf("Max_record_length %d\n", max_record_length);
	VM_printf("Max_fix_length %d\n", max_fix_length);
	VM_printf("Max_reset_length %d\n", max_reset_length);
    #endif
    if (!lopen(name)) exc_raise(LOG_CANNOT_OPEN);    
    /* active = LOG_FIX   --> info only in fix_table        */
    /* active = LOG_LOOP   --> info only in loop_table      */
    /* active = LOG_FIX|LOG_LOOP   --> info in both tables  */    
    if (loopSaveCount > 0) active |= LOG_LOOP;
    if (fixSaveCount > 0) active |= LOG_FIX;    
    if ( (loopSaveCount == 0) || 
	 (loop_start == start_time) || (loop_start == reset_time)) {	     
	     final_window = 1;
	     #ifdef __DEBUG__
	    	VM_printf("Setto final window all'inizio\n");
	     #endif
	 }    
    if (reset_time < MAX_DWORD) {
	/* There was a reset */
	if ((reset_start_mode == LOG_FIX)||(loop_start == reset_time) ) {
	    ptr = ((loop_start == reset_time) ? mem_table : reset_states);
	    grecord.t = reset_time;
	    grecord.act = ACT_UPDATE;
	    if (!lwrite(&grecord, SIZE)) exc_raise(LOG_CANNOT_WRITE);
	    if (!lwrite(ptr,MAX_PROC)) exc_raise(LOG_CANNOT_WRITE);
	} 
    }    

    while (active) {
	switch(active) {
	    case (LOG_LOOP | LOG_FIX) : /* Both queues are valid */ 
	    	if ( loop_table[loopScan].t <= fix_table[fixScan].t) {
		    if (!final_window) {
			#ifdef __DEBUG__
		    	VM_printf("Final window in (LOOP|FIX) - LOOP branch\n");
		    	#endif
		        grecord.t = loop_start;
			grecord.act = ACT_UPDATE;
			if (!lwrite(&grecord, SIZE))
			    exc_raise(LOG_CANNOT_WRITE);
		    	/* Store info to restore consistency 	*/
			/* at the beginning of the final window */
			if (!lwrite(&mem_table[0], MAX_PROC))
			    exc_raise(LOG_CANNOT_WRITE);
			final_window = 1;            
		    }
		    delta = log_save(loopScan, 1);
		    loopScan = (loopScan + delta) % loop_dim;
		    loopSaveCount -= delta;
		    if (loopSaveCount == 0) active &= ~LOG_LOOP;
		} else {
		    switch(fix_table[fixScan].act) {
			case ACT_START_WINDOW:
			    if (!final_window) {
				grecord.t = fix_table[fixScan].t;
				grecord.act = ACT_UPDATE;
				if (!lwrite(&grecord, SIZE)) 
				   exc_raise(LOG_CANNOT_WRITE);
				/* Store info to restore consistency */
				/* at the beginning of the window 	 */
				if (!lwrite(&fix_table[cfw * UPDATE_DIM].pid, MAX_PROC))
				    exc_raise(LOG_CANNOT_WRITE);
				fixSaveCount -= UPDATE_DIM;
				cfw++;
			    }
			    delta = 1;
			    break;
			case ACT_END_WINDOW:
			    if (final_window) delta = 1;
			    else if (fix_table[fixScan].t >= loop_start) {				
				#ifdef __DEBUG__
		    		VM_printf("Final window in (LOOP|FIX) - FIX-END branch\n");
			    	#endif
				final_window = 1;
				delta = 1;
			    }
			    else delta = log_save(fixScan, -1);
			    break;
			default:
			    delta = log_save(fixScan, -1);
			    break;
		    }
		    fixScan -= delta;
		    fixSaveCount -= delta;
		    if (fixSaveCount == 0) active &= ~LOG_FIX;
		}
		break;
	case LOG_LOOP : /* Only the Loop Queue is valid */
	    if (!final_window) {
		#ifdef __DEBUG__
		VM_printf("Final window in LOOP\n");
		#endif
		grecord.t = loop_start;
		grecord.act = ACT_UPDATE;
		if (!lwrite(&grecord, SIZE)) 
		    exc_raise(LOG_CANNOT_WRITE);
		/* Store info to restore consistency 	*/
		/* at the beginning of the final window */
		if (!lwrite(&mem_table[0], MAX_PROC)) 
		    exc_raise(LOG_CANNOT_WRITE);
		final_window = 1;            
	    }
	    delta = log_save(loopScan, 1);
	    last_clock = loop_table[loopScan].t;
	    loopScan = (loopScan + delta) % loop_dim;
	    loopSaveCount -= delta;
	    if (loopSaveCount == 0) active &= ~LOG_LOOP; 
	    break;
       case LOG_FIX :
       	    /* Only the fix queue is valid */
	    /* I'm sure I'm in the final window */
	    if ((fix_table[fixScan].act == ACT_START_WINDOW) ||
		(fix_table[fixScan].act == ACT_END_WINDOW) ) delta = 1;
	    else delta = log_save(fixScan, -1);
	    last_clock = fix_table[fixScan].t;
	    fixScan -= delta;
	    fixSaveCount -= delta;
	    if (fixSaveCount == 0) active &= ~LOG_FIX; 
	    break;
       default :
	    VM_printf("DUMP error 2\n"); 
	    exc_raise(LOG_DUMP_ERROR);
	    break;
	}
    }    
    /* Store info about tasks still alive  */
    grecord.t = last_clock + 10;
    grecord.act = ACT_INFO;
    for (i=0;i<MAX_PROC;i++)
        if ( cur_states[i] ) {
	    grecord.pid = i;
	    if (!lwrite(&grecord, SIZE)) exc_raise(LOG_CANNOT_WRITE);
	    memcpy(info.name, proc_table[i].name, NAME_LEN);
	    info.deadline = proc_table[i].drel * sys_tick;
	    info.type_class = (proc_table[i].type << 4) | proc_table[i].pclass;
	    info.period = proc_table[i].period * sys_tick;
	    info.creation = creation_times[i];
	    if (!lwrite(&info, sizeof(LOG_DATA)))
	        exc_raise(LOG_CANNOT_WRITE);
	}
	lclose();
        #ifdef __DEBUG__
	VM_printf("END dump info\n");
	VM_printf("Last clock = %d\n", last_clock);
	VM_printf("Final window  = %d\n", final_window);
	#endif
	if (fix_table != NULL) VM_free(fix_table, fix_dim * SIZE);
	VM_free(loop_table, loop_dim * SIZE);
}
