/* Project:     HARTIK 3.0                                      */
/* Description: Hard Real TIme Kernel for 8086 compatible       */
/* Author:      Giuseppe Lipari                                 */
/* Date:        2/7/96                                          */

/* File: 	Port.C						*/
/* Revision:    1.4						*/

#include <mem.h>
#include <string.h>

#include "vm.h"
#include "kern.h"
#include "port.h"

static char *port_error_msg[] = {
    "No more port descriptor","No more port interface",
    "No more hash entries","Port already open",
    "2 connection on same receive port",
    "Wrong port type","Unsupported port access",
    "Incompatible message declaration",
    "Wrong port operation","Port already closed"
};

static SEM hash_mutex;

#define __PORT_DBG__

struct hash_port {
	char	name[MAX_PORT_NAME];
	PID	port_index;
	BYTE 	blocked;
	SEM	sem;
	BYTE	valid;
	int	port_int;
};

struct port_ker {
	BYTE	valid;
	int 	next;
	WORD 	dim_block;
	BYTE	*mem_block;
	BYTE	*head;
	BYTE	*tail;
	BYTE	type;
	SEM	mutex;
	SEM	pieno;
	SEM	vuoto;
};

struct port_com {
	BYTE	valid;
	int 	next;
	PID 	proc_id;
	BYTE	access;
	int	port_index;
	WORD	dim_mes;
	int	h_index;
};
	
struct hash_port	htable[MAX_HASH_ENTRY];
struct port_ker		port_des[MAX_PORT];
struct port_com		port_int[MAX_PORT_INT];
QUEUE	freeportdes;
QUEUE	freeportint;

/*----------------------------------------------------------------------*/
/* Port default exception handlings					*/
/*----------------------------------------------------------------------*/
static int port_exception(int code)
{
    VM_printf("\nPort exception\n");
    VM_printf("Code [%d]\nCause : %s\n",code,port_error_msg[code-PORT_NO_MORE_DESCR]);
    /* This default exception handler will cause an abort */
    /* each time it is invoked				  */    
    return(0); 
}
 
/*----------------------------------------------------------------------*/
/* port_init() : inizializza le strutture delle porte; da chiamare 	*/
/* dalla sys_init().							*/
/*----------------------------------------------------------------------*/
void port_init(void)
{
    int i;

    /* Init hash table			*/
    for (i = 0; i < MAX_HASH_ENTRY; i++) {
	htable[i].valid = FALSE;
	htable[i].port_int = NIL;
	htable[i].blocked = 0;
    }

    /* mutex sem on the hash table		*/
    hash_mutex = sem_create(1);

    /* init the port descriptor table 	*/
    for (i = 0; i < MAX_PORT - 1; i++) {
	port_des[i].next = i+1;
	port_des[i].valid = FALSE;
    }
    port_des[MAX_PORT - 1].next = NIL;
    port_des[MAX_PORT - 1].valid = FALSE;
    freeportdes = 0;

    /* Init the port interface table	*/
    for (i = 0; i < MAX_PORT_INT - 1; i++) {
	port_int[i].next = i+1;
	port_int[i].valid = FALSE;
    }
    port_int[MAX_PORT_INT - 1].next = NIL;
    port_int[MAX_PORT_INT - 1].valid = FALSE;
    for (i = PORT_NO_MORE_DESCR; i <= PORT_UNVALID_DESCR; i++)
        exc_set(i,port_exception);
    freeportint = 0;
}


/*----------------------------------------------------------------------*/
/* hash_fun() : address hash table				 	*/
/*----------------------------------------------------------------------*/
static int hash_fun(char *name)
{
    return (*name % MAX_HASH_ENTRY);
}


/*----------------------------------------------------------------------*/
/* getfreedes : restituisce l'indice di un descrittore di porta libero	*/
/*----------------------------------------------------------------------*/
static int getfreedes(void)
{
    int p;

    VM_disable();
    if (freeportdes == NIL) {
	proc_table[exec].errno = PORT_NO_MORE_DESCR;
	VM_enable();
	return -1;
    }
    p = freeportdes;
    freeportdes = port_des[p].next;
    VM_enable();
    return(p);
}	  

/*----------------------------------------------------------------------*/
/* ungetdes() : mette il descrittore tra quelli disponibile	 	*/
/*----------------------------------------------------------------------*/
static void ungetdes(int pd)
{
    VM_disable();
    port_des[pd].next = freeportdes;
    freeportdes = pd;
    VM_enable();
}

/*----------------------------------------------------------------------*/
/* get freeint : restituisce una interfaccia di porta libera	 	*/
/*----------------------------------------------------------------------*/
static int getfreeint(void)
{
    int p;

    VM_disable();
    if (freeportint == NIL) {
	proc_table[exec].errno = PORT_NO_MORE_INTERF;
	VM_enable();
	return -1;
    }
    p = freeportint;
    freeportint = port_int[p].next;
    VM_enable();
    return(p);
}	

/*----------------------------------------------------------------------*/
/* ungetint : rende disponibile l'interfaccia di porta specificata	*/
/*----------------------------------------------------------------------*/
static void ungetint(int pi)
{
    VM_disable();
    port_int[pi].next = freeportint;
    freeportint = pi;
    VM_enable();
}

/*----------------------------------------------------------------------*/
/* port_create(): Apre la porta specificata dalla stringa, effettuando	*/
/* automaticamente il collegamento					*/
/* WARNING : La funzione e' bloccante per la mutua esclusione sulla 	*/
/* hash table	 						*/
/*----------------------------------------------------------------------*/
PORT port_create(char *name, int dim_mes, int num_mes, BYTE type, BYTE access)
{
    int i, pd, pi;
    WORD letti = 0;
    BYTE flag = FALSE;

/*
   Prendo un descrittore di porta.
*/
    #ifdef __PORT_DBG__
	if ((type == MAILBOX) && (access == WRITE)) {
	   proc_table[exec].errno = PORT_INCOMPAT_MESSAGE;
	   return -1;
	}
	if ((type == STICK  ) && (access == READ )) {
	   proc_table[exec].errno = PORT_INCOMPAT_MESSAGE;
	   return -1;
	}
    #endif
    pd = getfreedes();
    if (pd == -1) return -1;
/* 
   Devo andare in mutua esclusione sulla tabella di hash ! Poiche' questa
   viene acceduta dalle openport e dalle connect sono sicuro che una
   connect non puo' interrompere una openport.
*/
    sem_wait(hash_mutex,BLOCK);
/*
   Scorro la tabella di hash fino a trovare un'entrata libera, oppure 
   una entrata occupata in precedenza da una connect che ci si e' bloccata
   sopra (blocked > 0). Se ne trovo una gia' aperta da un'altra openport
   esco con errore. Lo stesso se sono state occupate tutte le entrate della
   tabella (tramite la var. letti);
*/
    i = hash_fun(name);
    while (!flag) {
	if (htable[i].valid == FALSE) flag = TRUE;
	else {
	    if (strcmp(htable[i].name,name) == 0) {
		if (htable[i].blocked == 0) {
		    proc_table[exec].errno = PORT_ALREADY_OPEN;
		    sem_signal(hash_mutex);
		    return -1;
		}
		else flag = TRUE;
	    }
	    else {
		i = (i+1) % MAX_HASH_ENTRY;
		letti++;
	    }
	}
	if (letti > MAX_HASH_ENTRY-1) {
	    proc_table[exec].errno = PORT_NO_MORE_HASHENTRY;
	    sem_signal(hash_mutex);
	    return -1;
	}

    }
    htable[i].valid = TRUE;
    strcpy(htable[i].name, name);

    htable[i].port_index = pd;
/* 
   A questo punto inizializzo tutta la struttura del descrittore di porta
*/
    if (type == STICK) 	port_des[pd].dim_block = dim_mes;
    else port_des[pd].dim_block = dim_mes * num_mes;

    VM_disable();
    port_des[pd].mem_block = VM_alloc(port_des[pd].dim_block);
    VM_enable();
    if (port_des[pd].mem_block == NULL) {
	proc_table[exec].errno = PORT_2_CONNECT;
	sem_signal(hash_mutex);
	return -1;
    }

    port_des[pd].head = port_des[pd].tail = port_des[pd].mem_block;

    port_des[pd].mutex = sem_create(1);
    port_des[pd].pieno = sem_create(port_des[pd].dim_block);
    port_des[pd].vuoto = sem_create(0);
    port_des[pd].type = type;
/*
   Prendo e inizializzo la struttura dell'interfaccia di porta verso il
   processo (e' la varibile pi quella che restituisco)
*/
    pi = getfreeint();
    if (pi == -1) {
	sem_signal(hash_mutex);
	return -1;
    }
	
/*  port_int[pi].proc_id = exec; */
    port_int[pi].access = access;
    port_int[pi].port_index = pd;
    port_int[pi].dim_mes = dim_mes;
    port_int[pi].next = NIL;
    port_int[pi].h_index = i;
    port_des[pd].valid = TRUE;
    port_int[pi].valid = TRUE;
/* 
   Sblocco eventuali processi che avevano fatto la connect nella coda
   semaforica che sta nell'entrata relativa della hash table !
*/
    if (htable[i].blocked > 0) {
	sem_xsignal(htable[i].sem, htable[i].blocked);
	htable[i].blocked = 0;
	sem_delete(htable[i].sem);
    }
/*
   Infine libero la mutua esclusione.
*/
    sem_signal(hash_mutex);
    return (pi);
}

/*----------------------------------------------------------------------*/
/* port_connect(): collega la porta specificata dalla stringa.		*/
/* WARNING : La funzione e' bloccante per la mutua esclusione sulle 	*/
/* strutture delle porte						*/
/*----------------------------------------------------------------------*/
PORT port_connect(char *name, int dim_mes, BYTE type, BYTE access)
{
    int i, pi, pd, pn,letti = 0;
    BYTE flag = FALSE, create = FALSE;

    #ifdef __PORT_DBG__
	if ((type == MAILBOX) && (access == READ )) {
	    proc_table[exec].errno = PORT_INCOMPAT_MESSAGE;
	    return -1;
	}
	if ((type == STICK  ) && (access == WRITE)) {
	   proc_table[exec].errno = PORT_INCOMPAT_MESSAGE;
	   return -1;
	}
    #endif
/*
   Per prima cosa mi prendo una interfaccia di porta e la riempio 
   parzialmente.	
*/
    pi = getfreeint();
    if (pi == -1) return -1;
/*  port_int[pi].proc_id = exec; */
    port_int[pi].access = access;
    port_int[pi].dim_mes = dim_mes;
    port_int[pi].next = NIL;
/*
   Mutua esclusione sulla tabella di hash
*/
    sem_wait(hash_mutex,BLOCK);
/*
   Cerco il descrittore appropriato : Se la porta e' gia' stata aperta
   allora esco dal ciclo con flag = TRUE, create = FALSE, l'indice i che 
   punta all'entrata della tabella di hash, e con htable[i].blocked = 0;
   In tutti gli altri casi significa che la porta non e' stata aperta e 
   quindi devo bloccarmi.
*/
    i = hash_fun(name);
    while (!flag) {
	/* Devo crearla */
	if (htable[i].valid == FALSE) {
	    flag = TRUE;
	    create = TRUE;
	}
	/* l'ho trovata (ma non so ancora se e' stata aperta) */
	else if (strcmp(htable[i].name, name) == 0) flag = TRUE;
	/* scandisco l'entrata successiva */
	else {
	    i = (i+1) % MAX_HASH_ENTRY;
	    letti ++;
	}
	#ifdef __PORT_DBG__
	    /* se ho scorso tutto l'array senza trovare nulla */
	    if (letti > MAX_HASH_ENTRY) {
		proc_table[exec].errno = PORT_NO_MORE_HASHENTRY;
		sem_signal(hash_mutex);
		return -1;
	    }
	#endif
    }
	
/*
   Se devo aspettare che venga aperta (create = TRUE) allora mi blocco in
   attesa sul semaforo della htable. Per non avere troppi semafori inutili
   in giro, me lo creo sul momento.
*/
    if (create == TRUE) {
	htable[i].valid = TRUE;
	htable[i].blocked = 1;
	strcpy(htable[i].name, name);
	htable[i].sem = sem_create(0);
	sem_signal(hash_mutex);
	sem_xwait(htable[i].sem,1,BLOCK);
    }
/*
   Se invece si e' gia' bloccata un'altra connect sopra, mi blocco anch'io.
   in ogni caso devo liberare la mutua esclusione sulla hash dato che ho
   gia' fatto tutto quello che dovevo fare con la hash.
*/
    else {
	if (htable[i].blocked > 0) {
	    htable[i].blocked++;
	    sem_signal(hash_mutex);
	    sem_xwait(htable[i].sem,1,BLOCK);
	}
	else sem_signal(hash_mutex);
    }
/*
   Controlli di errore.
*/
    pd = htable[i].port_index;
    #ifdef __PORT_DBG__
	if (type != port_des[pd].type) {
	    proc_table[exec].errno = PORT_UNSUPPORTED_ACC;
	    return -1;
	}
	if ((type == STICK) && (dim_mes != port_des[pd].dim_block)) {
	    proc_table[exec].errno = PORT_WRONG_OP;
	    return -1;
	}
	if ((type != STICK) && (port_des[pd].dim_block % dim_mes) != 0) {
	    proc_table[exec].errno = PORT_WRONG_OP;
	    return -1;
	}
    #endif

    sem_wait(hash_mutex,BLOCK);
    pn = htable[i].port_int;
    if (pn != NIL) {
	#ifdef __PORT_DBG__
	    if (type == STREAM) {
		proc_table[exec].errno = PORT_WRONG_TYPE;
		sem_signal(hash_mutex);
		return -1;
	    }
	    if (dim_mes != port_int[pn].dim_mes) {
		proc_table[exec].errno = PORT_WRONG_OP;
		sem_signal(hash_mutex);
		return -1;
	    }
	#endif
	port_int[pi].next = htable[i].port_int;
	htable[i].port_int = pi;
    }
    else htable[i].port_int = pi;
    sem_signal(hash_mutex);
    port_int[pi].h_index = i;
    port_int[pi].port_index = pd;
    port_int[pi].valid = TRUE;
    return(pi);
}

/*----------------------------------------------------------------------*/
/* port_delete() : inversa di port_open, libera tutto		 	*/
/*----------------------------------------------------------------------*/
void port_delete(PORT pi)
{
    int i;
    struct port_ker *pd;
    struct port_com *pp;

    pp = &port_int[pi];
    sem_wait(hash_mutex,BLOCK);
    i = pp->h_index;
    pd = &port_des[htable[i].port_index];
    pd->valid = FALSE;
    sem_delete(pd->mutex);
    sem_delete(pd->pieno);
    sem_delete(pd->vuoto);
    
    VM_disable();
    VM_free(pd->mem_block, pd->dim_block);
    VM_enable();

    ungetdes(htable[i].port_index);
    pp->valid = FALSE;
    htable[i].port_int = pp->next;
    ungetint(pi);
    htable[i].valid = FALSE;
    sem_signal(hash_mutex);
}

/*----------------------------------------------------------------------*/
/* port_disconnect() : libera l'interfaccia di porta 		 	*/
/*----------------------------------------------------------------------*/
void port_disconnect(PORT pi)
{
    sem_wait(hash_mutex,BLOCK);
    if (htable[port_int[pi].h_index].valid == TRUE)
        htable[port_int[pi].h_index].port_int = port_int[pi].next;
    port_int[pi].valid = FALSE;
    ungetint(pi);
    sem_signal(hash_mutex);
}

/*----------------------------------------------------------------------*/
/* port_send() : Invia un messaggio alla porta	 		 	*/
/*----------------------------------------------------------------------*/
WORD port_send(PORT pi, void *msg, BYTE wait)
{
    struct port_ker *pd;
    struct port_com *pp;

    pp = &(port_int[pi]);
    pd = &(port_des[pp->port_index]);
	
    #ifdef __PORT_DBG__
	if (pp->access == READ) exc_raise(PORT_WRONG_OP);
	if (!pd->valid) exc_raise(PORT_UNVALID_DESCR);
    #endif

    if (pd->type == STICK) sem_wait(pd->mutex,BLOCK);
    else if (pd->type == STREAM) {
            if (!sem_xwait(pd->pieno,pp->dim_mes,wait)) return(FALSE);
    }
    else {
	if (!sem_xwait(pd->pieno, pp->dim_mes,wait)) return(FALSE);
	sem_wait(pd->mutex,BLOCK);
    }

    memcpy(pd->head, msg, pp->dim_mes);

    pd->head += pp->dim_mes;
    
    if (pd->head >= (pd->mem_block + pd->dim_block)) 
	pd->head -= pd->dim_block;

    if (pd->type == STICK) sem_signal(pd->mutex);
    else if (pd->type == STREAM) sem_xsignal(pd->vuoto, pp->dim_mes);
    else {
	sem_xsignal(pd->vuoto, pp->dim_mes);
	sem_signal(pd->mutex);
    }
    return(TRUE);
}

/*----------------------------------------------------------------------*/
/* port_receive() : Riceve un messaggio dalla porta 		 	*/
/*----------------------------------------------------------------------*/
WORD port_receive(PORT pi,void *msg,BYTE wait)
{
    struct port_ker *pd;
    struct port_com *pp;

    pp = &(port_int[pi]);
    pd = &(port_des[pp->port_index]);
	
    #ifdef __PORT_DBG__
	if (pp->access == WRITE) exc_raise(PORT_WRONG_OP);
	if (!pd->valid) exc_raise(PORT_UNVALID_DESCR);
    #endif	

    if (pd->type == STICK) sem_wait(pd->mutex,BLOCK);
    else if (pd->type == STREAM) {
    	 if (!sem_xwait(pd->vuoto,pp->dim_mes,wait)) return(FALSE);
    }
    else {
	if (!sem_xwait(pd->vuoto,pp->dim_mes,wait)) return(FALSE);
	sem_wait(pd->mutex,BLOCK);
    }

    memcpy(msg, pd->tail, pp->dim_mes);

    pd->tail += pp->dim_mes;
    if (pd->tail >= (pd->mem_block + pd->dim_block)) 
	pd->tail -= pd->dim_block;

    if (pd->type == STICK) sem_signal(pd->mutex);	
    else if (pd->type == STREAM) sem_xsignal(pd->pieno, pp->dim_mes);
    else {
	sem_xsignal(pd->pieno, pp->dim_mes);
	sem_signal(pd->mutex);
    }
    return(TRUE);
}

#ifdef __PORT_DBG__

void print_port(void)
{
    int i;
    struct port_ker *pd;
    struct port_com *pp;
	
/*
	VM_printf("Hash Table :\n");
  	for (i=0; i<MAX_HASH_ENTRY; i++)
  		VM_printf("%d\tvl: %d\tbl: %d\tpd: %d\t%s\n", i,
  			htable[i].valid, htable[i].blocked, htable[i].port_index,
  			htable[i].name);
*/
	VM_printf("Port des :\n");
	VM_printf("Free port des : %d\n", freeportdes);
	for (i=0; i<MAX_PORT_INT; i++)
	    if (port_int[i].valid) {
		pp = &port_int[i];
		pd = &port_des[pp->port_index];
    		VM_printf("%d %s  vt: %d      pn: %d\n",i,htable[pp->h_index].name,
        	pd->vuoto,pd->pieno);
	    }
/*
	VM_printf("Port int :\n");
  	VM_printf("Free port int : %d\n", freeportint);
  	for (i=0; i<MAX_PORT_INT; i++)
  		VM_printf("%d  vl : %d  dm : %d  port_index : %d  proc_id : %d\n", i, 
  			port_int[i].valid, port_int[i].dim_mes, 
  			port_int[i].port_index, port_int[i].proc_id);
*/
}

void port_write(PORT p)
{
    struct port_ker *pd;
    struct port_com *pp;
    char msg[80];

    pp = &port_int[p];
    pd = &port_des[pp->port_index];

    VM_printf(msg,"%d  pd: %d   vt: %d      pn: %d ",p,pp->port_index,
        pd->vuoto,pd->pieno);
}


int isBlocked(PID i)
{
    if (proc_table[i].status == WAIT) return (TRUE);
    else return (FALSE);
}

#endif
