/* Project:     HARTIK 3.0                                      */
/* Description: Hard Real TIme Kernel for 8086 compatible       */
/* Author:      Giuseppe Lipari                                 */
/* Date:        1/10/96                                         */

/* File:        Sem.C                                           */
/* Revision:    1.0                                             */

/* Semaphores:                                                          */
/* this is the generalized version of the primitives signal & wait      */
/* In this case, the user can specify the number to inc/dec the         */
/* semaphore's counter. It is useful in the buffer management           */
/* (see port section)                                                   */

#include <string.h>
#include <stdlib.h>

#include "vm.h"
#include "kern.h"

#ifdef __TRACE__
#include "log.h"
#endif

/* This file contains all the standard semaphores primitives    */

/* Semaphores descriptor tables */
static struct sem_des {
    int count;
    int q_first;
    int q_last;
    QUEUE next;
    BYTE used;
} sem_table[MAX_SEM];

static struct sproc {
    int decsem;
    int next;
} sp_table[MAX_PROC];

QUEUE blocked;          /* Blocked queue                        */
QUEUE free_sem;         /* Queue of free sem                    */

extern DWORD last_int;		/* Time of the last system call
			   (in usec) --- Luca 			*/
extern DWORD act_time;		/* Actual time, readed by every system
			   call that manages the capacity...
			   It's global because the task_makeready
			   problem...				*/

			   /* 4 DBG...*/ extern int ldbg;

/*----------------------------------------------------------------------*/
/* Init the semaphoric structures                                       */
/*----------------------------------------------------------------------*/
void sem_init(void)
{
    unsigned i;
    for (i = 0; i < MAX_SEM; i++) {
	sem_table[i].used = FALSE;
	sem_table[i].next = i+1;
	sem_table[i].q_first = NIL;
	sem_table[i].q_last = NIL;
    }
    sem_table[MAX_SEM-1].next = NIL;
    free_sem = 0;
    for (i = 0; i < MAX_PROC; i++) sp_table[i].next = NIL;
    blocked = NIL;
}

/*----------------------------------------------------------------------*/
/* Allocate a semaphoric descriptor and sets the counter to n           */
/*----------------------------------------------------------------------*/
SEM sem_create(int n)
{
    SEM q;
    
    VM_disable();
    q = free_sem;
    if (q != NIL) {
	free_sem = sem_table[q].next;
	sem_table[q].count = n;
	sem_table[q].q_first = NIL;
	sem_table[q].q_last = NIL;
	sem_table[q].used = TRUE;
    }
    else {
	proc_table[exec].errno = NO_MORE_SEM;
	VM_enable();
	return -1;
    }
    #ifdef __HOOK__ 
	act_exc(SEM_CREATE_HOOK);
    #endif    
    VM_enable();
    return(q);
}

/*----------------------------------------------------------------------*/
/* Frees a semaphores descriptor                                        */
/*----------------------------------------------------------------------*/
void sem_delete(SEM q)
{
    VM_disable();
    sem_table[q].used = FALSE;
    sem_table[q].next = free_sem;
    free_sem = q;
    #ifdef __HOOK__ 
	act_exc(SEM_DELETE_HOOK);
    #endif    
    VM_enable();
}

/*----------------------------------------------------------------------*/
/* Generic wait. If it is possible, decrements the sem counter of n,    */
/* else blocks the task.                                                */
/*----------------------------------------------------------------------*/
WORD sem_xwait(SEM s, int n, BYTE wait)
{
    struct sem_des *s1; /* It speeds up access */
    /* First, we have to manage the capacity of the running task...	*/
    /* 		----By Luca	:-(					*/

    CBS_PROLOGUE();
    /* We do not need to save context if we are sure we shall not block! */
    if (wait == NON_BLOCK) VM_disable();    
    else proc_table[exec].context = VM_context_save();
    s1 = &sem_table[s];    
    if (s1->used != TRUE) act_exc(UNVALID_SEM);
    /* The non blocking wait is really simple! */
    /* We do not suspend or schedule anything  */
    if (wait == NON_BLOCK) {
	if (s1->count < n) {
	    #ifdef __HOOK__ 
		act_exc(SEM_WAIT_HOOK);
	    #endif
	    #ifdef __TRACE__
		log_record(exec,ACT_WAITNB);
	    #endif
	    VM_enable();
	    return(FALSE);
	}
	else s1->count -= n;
	#ifdef __HOOK__ 
	    act_exc(SEM_WAIT_HOOK);
	#endif
	#ifdef __TRACE__
	    log_record(exec,ACT_WAITNB);
	#endif
	VM_enable();
	return(TRUE);
    }
    /* The blocking wait is more complex... */
    else {
	if (s1->count < n)  {           /* We must block exec task   */
	    proc_table[exec].status = WAIT;
	    q_insert(exec,&blocked);
	    #ifdef __TRACE__
		log_record(exec,ACT_WAIT);
	    #endif
	    sp_table[exec].decsem = n;  /* Prepare sproc des...       */
	    sp_table[exec].next = NIL;
	    if (s1->q_last != NIL)      /* ...and put it in sem queue */
		sp_table[s1->q_last].next = exec;
	    else    s1->q_first = exec;
	    s1->q_last = exec;
	    dispatch();
	}
	else {
	    s1->count -= n;
	    #ifdef __TRACE__
		log_record(exec,ACT_WAIT);
	    #endif
	}
	#ifdef __HOOK__ 
	    act_exc(SEM_WAIT_HOOK);
	#endif
	VM_context_change(proc_table[exec].context);
    }
    return(TRUE);
}

/*----------------------------------------------------------------------*/
/* Generic signal. It increments the sem counter of n, and wakes all the*/
/* tasks that are blocked on the semaphore, if it is possible. The      */
/* semaphoric queue is a FIFO queue, in order to eliminate deadlocks    */
/*----------------------------------------------------------------------*/
void sem_xsignal(SEM s, int n)
{
    struct sem_des *s1;        /* it speeds up access          */
    int p;                     /* idem                         */
    BYTE fl;
    #ifdef __TRACE__
	BYTE fl1 = FALSE;
    #endif

    proc_table[exec].context = VM_context_save();
    /* First, we have to manage the capacity of the running task...	*/
    /* 		----By Luca	:-(					*/

    CBS_PROLOGUE();

    s1 = &sem_table[s];
    if (s1->used != TRUE) act_exc(UNVALID_SEM);
    s1->count += n;                     /* inc sem count                */
    p = s1->q_first;

    fl = (p != NIL);
    while (fl) {
	if (proc_table[p].status == WAIT) {
	    if (sp_table[p].decsem <= s1->count) {
		/* Dec sem count */
		s1->count -= sp_table[p].decsem;
		/*Get task from blocked queue */
		q_extract(p,&blocked);         
		proc_table[p].status = READY;
		#ifdef __TRACE__
		    fl1 = TRUE;
		    log_record(p,ACT_SIGNAL);
		#endif
		/* Put it in ready queue    */
		if (proc_table[p].pclass == HARD) {
		    q_insert(p,&ready_rt);
		} else if (proc_table[p].pclass == SOFT) {
		    /* When a new CBS task arrives to the system, we must check the
		    residual CPU band of the CBS... */
		    if (proc_table[p].dline * sys_tick < act_time +
			(proc_table[p].cap * proc_table[p].period * sys_tick) / proc_table[p].q){
			    proc_table[p].cap = proc_table[p].q;
			    proc_table[p].request = act_time / sys_tick;
			    proc_table[p].dline = proc_table[p].request + proc_table[p].period;
			    proc_table[p].priority = proc_table[p].dline;
		    }
		    q_insert(p,&ready_rt);
		} else q_insert(p,&ready_nrt);
		/* Get task p from sem queue    */
		if (s1->q_last == p) s1->q_last = NIL;
		s1->q_first = sp_table[p].next;
		/* Next task to wake            */
		p = s1->q_first;
		fl = (p != NIL);
	    }
	    /* This task CANNOT be awaken, so exit the loop */
	    else fl = FALSE;	    
	}
	else {
	    /* This task was killed and we must NOT dec the counter     */
	    /* get task p from sem queue                                */
	    if (s1->q_last == p) s1->q_last = NIL;
	    s1->q_first = sp_table[p].next;
	    /* Next task to wake                                        */
	    p = s1->q_first;
	    fl = (p != NIL);
	}
    }
    #ifdef __TRACE__
	if (!fl1) log_record(exec,ACT_SIGNAL);
    #endif
    scheduler(); /* Do preemption if needed      */
    #ifdef __HOOK__ 
	act_exc(SEM_SIGNAL_HOOK);
    #endif    
    VM_context_change(proc_table[exec].context);
}

