/* Project:     HARTIK 3.0                                      */
/* Description: Hard Real TIme Kernel for 8086 compatible       */
/* Author:      Gerardo Lamastra                                */
/*              Giuseppe Lipari                                 */
/* Date:        30/9/96                                         */

/* File:        SRP.C                                           */
/* Revision:    1.0                                             */

/* This file contains all the kernel level functions and most   */
/* of basic task primitives.                                    */
/* A brief description of the SRP mechanism:
 * SRP is a very simple resource control protocol, it is based
 * on the idea of giving to each resource a static ceiling, which
 * is the shortest relative deadline of the task using the resource;
 * The system keep track of the current ceiling defined as the minimum
 * ceiling among the resource which are currently held by a task.
 * Then a task can be scheduled only if its relative deadline is
 * strictly less than the system ceiling.
 * This basically means to avoid preemption from equal or less priority
 * task when a low priority task is holding a resource.
 * Implementation: in order to implement SRP, each time a task is created
 * we need to update the ceiling of the resources it uses; then when a task
 * enters a resource, the system ceiling is stored on a stack, then the
 * ceiling is updated as the min(sys_ceiling,res_ceiling[used])
 * When the resource is released, the old ceiling is popped from the stack
 * Guarantee: in order to support dynamic task creation/distruction, the
 * ceiling values MUST be updated only when no task is holding a critical
 * section; so if the ceiling stack is non empty, this means that a task
 * is inside a critical section; then if a new task is created, ceiling 
 * values cannot be changed; again if a task just created is activated
 * the ceiling values cannot be changed. A task is hold in the LOBBY or
 * LOBBY_READY state for this purpose; when the stack is empty, no resource
 * is used and the ceiling may be updated. This usually happens when a
 * signal is issued. Of course, if no resource is hold, ceiling values
 * may be immediately updated!
 */

#include <mem.h>
#include <string.h>
#include <stdlib.h>

#include "vm.h"
#include "kern.h"
#include "srp.h"

static DWORD srp_stack[MAX_RESOURCE];
static int srp_tos;

DWORD res_ceiling[MAX_RESOURCE];
DWORD new_ceiling[MAX_RESOURCE];
SRP_ENTRY srp_block[MAX_PROC];
PID srp_first;
DWORD sys_ceiling = MAX_CEILING;
QUEUE lobby;
QUEUE ending;

/* Init the SRP system tables */
void srp_init(void)
{
    register r;

    srp_first = NIL;
    srp_tos = 0;
    lobby = NIL;
    ending = NIL;
    sys_ceiling = MAX_DWORD;
    
    for (r = 0; r < MAX_RESOURCE; r++) 
	res_ceiling[r] = new_ceiling[r] = MAX_DWORD;
    for (r = 0; r < MAX_PROC; r++) {
	srp_block[r].nxt = NIL;
	memset(srp_block[r].section,0,sizeof(DWORD)*MAX_RESOURCE);
    }
}

/* Do ceiling calculation! */
void srp_calc_ceiling(DWORD *tmp_ceiling)
{
    register r,p;

    for (r = 0; r < MAX_RESOURCE; r++) {
	tmp_ceiling[r] = MAX_DWORD;
	p = srp_first;
	while ((p != NIL) && (tmp_ceiling[r] == MAX_DWORD)) {
	    if (srp_block[p].section[r] == 0) p = srp_block[p].nxt;
	    else tmp_ceiling[r] = proc_table[p].period;
	}
    }
}

/* SRP Stack management! */
void srp_push(DWORD r)
{
    if (srp_tos == MAX_RESOURCE) act_exc(SRP_STACK_FULL);
    else srp_stack[srp_tos++] = r;
}

DWORD srp_pop(void)
{
    if (srp_tos == 0) act_exc(SRP_STACK_EMPTY);
    return(srp_stack[--srp_tos]);
}

int srp_stack_empty(void)
{
    if (srp_tos == 0) return 1;
    else return 0;
}

/* SRP task table management; used for storing all the task which
 * use resources
 */
void srp_insert(PID i)
{
    PID p,q;

    p = NIL;
    q = srp_first;
    while (q != NIL && (proc_table[i].drel > proc_table[q].drel)) {
	p = q;
	q = srp_block[q].nxt;
    }
    if (p == NIL) {
	srp_block[i].nxt = srp_first;
	srp_first = i;
    }
    else {
	srp_block[p].nxt = i;
	srp_block[i].nxt = q;
    }
}

void srp_extract(PID p)
{    
    PID q1;

    q1 = srp_first;
    if (q1 == p) {
	srp_first = srp_block[p].nxt;
	return;
    }
    
    else {
	while (q1 != NIL && srp_block[q1].nxt != p)
	    q1 = srp_block[q1].nxt;
	if (q1 == NIL) act_exc(SRP_BAD_EXTRACT);
	srp_block[q1].nxt = srp_block[p].nxt;
	return;
    }
}

/* Wait/signal function */
void srp_wait(int resource)
{
    VM_disable();
    srp_push(sys_ceiling);
    sys_ceiling = min(sys_ceiling,res_ceiling[resource]);
    VM_enable();
}

/* Resource is only used for test purpose!!
*/
void srp_signal(int resource)
{
    PID q;
    
    proc_table[exec].context = VM_context_save();
    sys_ceiling = srp_pop();
    if (srp_stack_empty()) {
	if (lobby != NIL)
	   memcpy(res_ceiling,new_ceiling,sizeof(DWORD)*MAX_RESOURCE);
	while (lobby != NIL) {
	    q = q_getfirst(&lobby);
	    if (proc_table[q].status == LOBBY) proc_table[q].status = SLEEP;
	    else {
		proc_table[q].request = sys_clock;
		proc_table[q].dline = sys_clock + proc_table[q].drel;
		proc_table[q].priority = proc_table[q].dline;
		proc_table[q].status = READY;
		q_insert(q,&ready_rt);
	    }
	}
	if (ending != NIL) {
	    while (ending != NIL) {
		q = q_getfirst(&ending);
		proc_table[q].status = FREE;
		q_insert(q,&freedesc);
		/* not needed */
		//srp_extract(q);
	    }
	    srp_calc_ceiling(res_ceiling);
	    memcpy(new_ceiling,res_ceiling,sizeof(DWORD)*MAX_RESOURCE);
	}
    }
    scheduler();
    VM_context_change(proc_table[exec].context);
}

/*----------------------------------------------------------------------*/
/* Guarantee an hard task with SRP resource management.                 */
/* Look at theorical stuff into Peppe's tesi !                          */
/*----------------------------------------------------------------------*/
#define __USE_GUARANTEE__
static int __guarantee(void)
{
    DWORD tmp_ceiling[MAX_RESOURCE];
    DWORD bp[MAX_PROC];
    register r;
    PID p,q;
    double util = 0;
    double utemp = 0;

#ifdef __USE_GUARANTEE__
    srp_calc_ceiling(tmp_ceiling);
    /* Calculate the Blocking time for hard task */
    p = srp_first;
    while (p != NIL) {
	bp[p] = 0;
	q = srp_block[p].nxt;
	while (q != NIL) {
	    for (r = 0; r < MAX_RESOURCE; r++) {
		if (tmp_ceiling[r] <= proc_table[p].period) {
		     bp[p] = max(bp[p],srp_block[q].section[r]);
		}
	    }
	    q = srp_block[q].nxt;
	}
	p = srp_block[p].nxt;
    }

    /* Guarantee ! */
    p = srp_first;
    util = 0;
    while (p != NIL) {
#ifdef __TBS__	
	if (proc_table[p].pclass != SOFT) {
#endif
	    util += ((double)proc_table[p].wcet)/
	    	    ((double)(proc_table[p].period * sys_tick));
	    utemp = util + ((double)bp[p])/
	    	((double)(proc_table[p].period * sys_tick));
#ifdef __TBS__	    
	    utemp += (((double)(TBS_util))/(double)(TBS_PREC));
#endif	    
	    if (utemp > 1) return (0);
	    else p = srp_block[p].nxt;
#ifdef __TBS__	    
	}
	else p = srp_block[p].nxt;
#endif	
    }

    /* If guarantee, set the new ceiling vector */
    memcpy(new_ceiling,tmp_ceiling,sizeof(DWORD)*MAX_RESOURCE);
    return(1);
#else
    srp_calc_ceiling(new_ceiling);
    return(1);
#endif    
}

int srp_guarantee(PID i)
{
    VM_FPU_save();
    /* Put the task in the srp table */
    srp_insert(i);
    if (!__guarantee()) {
	srp_extract(i);
	VM_FPU_restore();
	return -1;
    } else srp_starttask(i);
    VM_FPU_restore();
    return 1;
}

#ifdef __SRP__

void srp_endtask(PID p)
{
    srp_extract(p);
    if (!srp_stack_empty()) {
	proc_table[p].status = ENDING;
	q_insert(p,&ending);
    } else {
	srp_calc_ceiling(res_ceiling);
	/* For consistency we keep
	 * new_ceiling = res_ceiling whenever possible...
	 */
	memcpy(new_ceiling,res_ceiling,sizeof(DWORD)*MAX_RESOURCE);
	proc_table[p].status = FREE;
	q_insert(p,&freedesc);
    }
}

void srp_starttask(PID p)
{
    if (!srp_stack_empty()) {
	proc_table[p].status = LOBBY;
	q_insert(p,&lobby);
    } else {
	memcpy(res_ceiling,new_ceiling,sizeof(DWORD)*MAX_RESOURCE);
    }
}

#endif
