/* Project:     HARTIK 3.0                                      */
/* Description: Hard Real TIme Kernel for 8086 compatible       */
/* Author:      Gerardo Lamastra                                */
/* Date:        12/7/98                                         */
/* Revision:    Beta 2.0                                        */

/* File: 	VmMem.C						*/
/* Basic Memory Manager Functions:
   This is essentially based onto the K&R allocator with the
   approach followed by Antonino Casile in the DEC Alpha
   Virtual Machine implementation!
*/
#include <mem.h>
#include <string.h>

#include "vm.h"
#include "kern.h"

/* L'allocazione di memoria avviene in multipli di una unita' atomica
 * detta BLOCCO ( la cui lunghezza e' stata fissata per ora in 32 bytes ) .
 * Questo approccio consente di fare , all'atto del rilascio della 
 * memoria , una fase di garbage collection la cui durata in termini di
 * tempo risulta limitata ( cosa esenziale in un sistema Hard Real-Time).
 */
/* BLOCK_BITS e' il numero di bit di offset all'interno di un blocco .
 * Si noti come questo limita le lunghezze del blocco alle potenze di 2 ,
 * cosa che mi pare perfettamente logica e non limitativa .
 */

/* Non-significat bits in a BLOCK */
#define BLOCK_BITS	5  
/* Basic BLOCK length in bytes */
#define BLOCK_SIZE	(1L << BLOCK_BITS)

typedef struct mem_chunk {
    unsigned long int size; /* Lenght in BLOCK units of this chunk */
    struct mem_chunk *next; /* Next free chunk pointer */
} MEM_CHUNK;
	
/* This is the first free CHUNK */	
static MEM_CHUNK *first_chunk;

/* tot_blocks gives the number of free blocks */
static int act_blocks;

void VM_mem_init(void *base, DWORD size)
{
    first_chunk = base;
    first_chunk->size = act_blocks = size >> BLOCK_BITS;
    first_chunk->next = NULL;
}

void *VM_alloc(DWORD size)
{
    int p_size = ((size + BLOCK_SIZE - 1) >> BLOCK_BITS);
    MEM_CHUNK *temp = first_chunk;
    MEM_CHUNK *temp1 = NULL;
    MEM_CHUNK *ret_addr;
    
    if (act_blocks < size || temp == (MEM_CHUNK *)NULL) {
	return(NULL);
    }
    while (temp->size < p_size && temp->next != NULL) {
	temp1 = temp;
	temp = temp->next;
    }    
    if (temp->size < p_size && temp->next == NULL) {
	/*
	Memory fragmentation!! Even if the sum all the free
	chunks is greater or equal than p_size, we cannot find
	a chunk which is big enough to satisfy the request
	*/
	return(NULL);
    }
    ret_addr = temp;
    act_blocks -= p_size;    
    temp->size -= p_size;

    if (temp->size == 0) {
	if (temp != first_chunk) temp1->next = temp->next;
	else first_chunk = temp->next;
    } else {
	if (temp != first_chunk) temp1->next += p_size;
	else first_chunk = p_size;
    }
    return(ret_addr);
}

WORD VM_free(void *block, DWORD size)
{
    DWORD p_size = ((size + BLOCK_SIZE - 1) >> BLOCK_BITS);
    MEM_CHUNK *temp;
    MEM_CHUNK *chunk = (MEM_CHUNK *)block;
    BYTE merge = FALSE;
     
    if (first_chunk == NULL) {
	chunk->size = p_size;
	chunk->next = NULL;
	first_chunk = chunk;
    } else if (first_chunk > chunk) {
	/* NOTA BENE : Prima di fare il calcolo ho dovuto fare il casting
	 * a void * poiche' se no altrimenti il compilatore fa automatica-
	 * mente puntatore+num*sizeof(struttura a cui punta puntatore ) .
	 * Idem per le altre ricorrenze di questo casting .
	*/
	if ((BYTE *)(chunk+p_size*BLOCK_SIZE) == (BYTE *)first_chunk) {
	    chunk->next = first_chunk->next;
	    chunk->size = p_size + first_chunk->size;
	    first_chunk = chunk;
	} else {
	    chunk->next = first_chunk;
	    chunk->size = p_size;
	    first_chunk = chunk;
	}
    } else {
	temp = first_chunk;
	while (temp->next < chunk && temp->next != NULL) temp = temp->next;
	if ((BYTE *)temp+(BLOCK_SIZE*temp->size) == (BYTE *)(chunk)) {	    
	    /* Entrare in questo blocco vuole dire che il
	     * blocco rilasciato puo' essere fuso con il
	     * blocco che lo precede.
	    */
	    temp->size += p_size;
	    chunk = temp;
	    p_size = temp->size;
	    merge = TRUE;
	}
	if ( (BYTE *)chunk+p_size*BLOCK_SIZE == (BYTE *)(temp->next)) {
	    /* Entrare in questo blocco vuole dire che il
	     * blocco rilasciato puo' essere fuso con il
	     * blocco che lo segue .
	    */
	    chunk->size = p_size+(temp->next)->size;
	    chunk->next = (temp->next)->next;
	    if (merge != TRUE) temp->next = chunk;
	} else if (merge == FALSE) {
	    /* Entrare in questo blocco vuole dire che il
	     * blocco rilasciato non puo' essere fuso con 
	     * nessun altro blocco e quindi viene solo
	     * accodato .
	    */
	    chunk->next = temp->next;
	    chunk->size = p_size;
	    temp->next = chunk;
	}
    }
    act_blocks += (size+BLOCK_SIZE-1)>>BLOCK_BITS;
    return(1);
}

void VM_mem_dump(void)
{
    MEM_CHUNK *temp = first_chunk;
    register i = 0;

    do {
	VM_printf("Chunk %d --> Address %p | Next %p | Size %lu\n",
		i,temp,temp->next,temp->size);
 	temp = temp->next;
 	i++;
    } while (temp != NULL);
    VM_printf("Act_Blocks = %ld\n",act_blocks);
}


