/* Project:     HARTIK 3.0                                      */
/* Description: Hard Real TIme Kernel for 386 & higher machines */
/* Author:      Gerardo Lamastra                                */
/* Date:        9/5/96                                          */
/* Revision:    Beta 2.0                                        */

/* CVS $Id: x_stub.c,v 1.3 1999/05/21 10:57:40 massy Exp $ */

/* Module: X_Stub.C                                              */
/* main() & system tables management functions                   */

/* The X extender provides the basic services necessary to load  */
/* and run a PM application; we do not need the whole DPMI stuff */
/* so we have implemented just some basic functionality:         */
/* 1) Mode switch handling                                       */
/* 2) Memory allocation/deallocation using XMS services or raw   */
/*    A20 interface (PS2 & Classic AT supported!)                */
/* 3) Memory transfer/compare/set enhanced to manage extended    */
/*    memory                                                     */
/* 4) .ELF loading & execution                                   */

/*
   VERSION HISTORY
   ------- -------
   Beta 1.0 : First organization of a set of indipendent experiment about
	      mode switching, xms management, ELF loading support; it seems
	      to work correctly; just tested with a simple file; necessary
	      to develop independent library for I/O functions.

   Beta 1.1 : Fixing a problem with register passing from RM to PM
	      the WASM uses the SS register to access variables into
	      data segment and this caused some minor problem.
	      (File X_Switch.Asm)

   Beta 1.2 : Added info struct to allow command lime parameters passing
	      and info about memory
    
   Beta 1.3 : Experimented the same problem in passing registers from
	      PM to RM when the application ended; the cause was the same
	      as in the previous problem.
	      Fixed a problem cause by BP corrupting during the switch!
	      (File X_Switch.Asm)

   Beta 1.4 : Resolved a little problem when using task switching hardware
	      due to the TS bit into CR0; the bit was left active when
	      switching back to real mode; this would cause an exception
	      when another protected application switch to PM.
	      (File X_Switch.Asm)

   Beta 1.5 : Fixed a very obscure bug due to elf loading; when we load
	      the ELF executable we calculated needed memory summing up the
	      p_memsize field of each segment & using p_align to ensure
	      correct alignment; this revealed not completely correct as
	      each segment MUST reside in memory at the same virtual address
	      at which the linker allocated it; this can result into greater
	      memory needing depending on virtual address settings!
	      (File X_Elf.C)

   Beta 2.0 : Adding support for BIOS calling service from PM
	      (File X_Switch.Asm,X_Stub.C)

   Beta 2.2 : Keyboard is reset upon exit, like mouse (X_Stub.C)

   Beta 2.3 : When XMS was not active, the A20_query procedure which
	      used the GS/FS register would hang the system; the X_A20.Asm
	      module has been made fully 8086 compatible to avoid this
	      problem. I cannot figure out why this happened; it seems
	      that certain mainboards are more subject to crash, especially
	      the most powered!
	      
   Beta 3.0 : Added COFF loading support              

   Beta 3.1 : Adjusted COFF loading support; now we use static buffers and
	      direct I/O functions

   Beta 3.2 : I finally discovered another bug! GDT size was set to 256
	      entries; anyway we must pay a lot of attention to this.
	      If we insert a selector whose index is greater than 0x100
	      this would  be outside of GDT allocated space.
	      As we assign selectors from 0x100 to 0x310 (Using 66 tasks)
	      this would cause unpredictable results!
	      Now the GDT size is set to 1024 entries; this should be
	      enough. (I hope!)

   Beta 3.3   The Watcom Linker 11.0 produce a slightly different
	      ELF executable; I have observed two modifications:
	      1) We now need to specify with OPTION OFFSET=0
		 that we want that the application virtual address space
		 starts from 0x000000; Otherwise the linker produce a
		 different value, usually 0x80
	      2) The linker generates another Segment, which is of
		 PT_PHDR type; pratically it is a copy into the disk image
		 of the Program Header Table; It seems that this is used
		 for program interpreting... II just ignored It
	      Refer to X_Elf.C to analyze further the modification I have
	      done!

   Beta 3.4   Fixed a minor problem when accessing to High Memory with
	      himem.sys not active... Also, added the __XTN__ symbol
	      when compiling, to define that the eXtender is being
	      processed. This turns to be useful in elf\hw386.h when
	      defining some 32 bit inline funcs...

   Beta 3.5   Removed calling mouse_reset() and keyb_reset(). The reset
              is made by the drivers under Hartik.
	      By MG.
	      
*/

#define __X_VERSION__   0x0305

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <conio.h>
#include <io.h>
#include <i86.h>

#include "x86\hw386.h"
#include "x_elf.h"
#include "x_coff.h"
#include "x_mem.h"
#include "x_stub.h"

/* These are the system tables */

union gdt_entry GDT[256];
GATE IDT[256];

/* Save here the real mode CS & DS/SS/ES                 */
/* As we use a small model to build XTN we only need to  */
/* know the value of CS & ES; SS & ES are the same as DS */

WORD real_code,real_data;

/* These variables contains the RM masks for interrupt controllers */
/* The PIC status is restored when the PM functions returns to RM  */

BYTE real_pic1_mask;
BYTE real_pic2_mask;

/* X will send useful info using this structure to PM application */
struct info X_info;

/* Real mode pointers to PM code functions                              */
/* This pointers are used to set up the 32 bit code segment suitable    */
/* for executing 32 bit extender functions such as mem... functions     */
/* and to set up the gate used to return to real mode from PM           */
/* I will use the following convetion: if a function foo() jumps to     */
/* a protected mode stub, I will call the PM stub _foo()                */

extern void __far __cdecl _x_call32(void);
void (__far * __far __cdecl _x_call32_ptr)(void) = _x_call32;
extern void __far __cdecl _x_ret16(void);
void (__far * __far __cdecl _x_ret16_ptr)(void) = _x_ret16;
extern void __far __cdecl _x_callBIOS(void);
void (__far * __far __cdecl _x_callBIOS_ptr)(void) = _x_callBIOS;

/* This bunch of memory is reserved as protected mode stack  */
/* We need to have at least two different stack, one for RM  */
/* and another for PM; the RM stack is of course provided    */
/* with C startup code; the PM stack is defined here; anyway */
/* the spawned PM application should set up its own stack    */

#define PM_STACK_SIZE   0x100
BYTE PM_stack[PM_STACK_SIZE];
DWORD PM_esp;

/* This is the RM loading buffer */

BYTE load_buf[BUFSIZE];

/* Translate real mode address into linear address */
/* This function is not implemented as macro       */
/* to avoid strange side-effects                   */
/* Linear address is SEG*16+OFFSET                 */

DWORD LINEARIZE(void __far *address)
{
    DWORD result;
    result = FP_SEG(address);
    result <<= 4;
    result += FP_OFF(address);
    return(result);
}

/* A gate can be inserted into GDT or IDT */
/* Refer to Hw386.H for gate structure    */

void ins_gate(WORD id,void *where,WORD s,DWORD o,BYTE a,BYTE wc)
{
    GATE *g;
    id >>= 3;
    g = ((GATE *)(where));
    g = &g[id];
    g->sel = s;
    g->access = a;
    g->dword_cnt= wc;
    g->offset_lo = (o & 0xFFFF);
    o >>= 16;
    g->offset_hi = (o & 0xFFFF);
}

/* A descriptor can only be inserted into GDT */
/* Refer to Hw386.H for gate structure        */

void ins_descr(WORD id,DWORD b,DWORD l,BYTE a,BYTE g)
{
    id >>= 3;
    GDT[id].d.base_lo = (b & 0x0000FFFF);
    GDT[id].d.base_med = ((b & 0x00FF0000) >> 16);
    GDT[id].d.base_hi = ((b & 0xFF000000) >> 24);
    GDT[id].d.access = a;
    GDT[id].d.lim_lo = (l & 0xFFFF);
    GDT[id].d.gran = (g | ((l >> 16) & 0x0F));
}

/* This function initialize the GDT entries & other useful variables       */
/* The GDT contains different groups of selectors: NULL_SEL is of course   */
/* the null selector, representing a sort of null pointer; the X_VIDEO_SEL */
/* is used to access video memory, mainly for debugging purposes in this   */
/* module; the X_FLAT... selectors are used for accessing flat memory;     */
/* the X_CODE16_SEL,X_CODE32_SEL & X_DATA_SEL are used to map the RM & PM  */
/* code for X; X_DATA_SEL is used to access X data variables from PM; it   */
/* is also useful when switching back to RM in order to reload segment     */
/* data registers with values suitable for RM; finally the X_RM_BACK_GATE  */
/* is used as an handle to come back to RM from any PM code segment also   */
/* if this segment is not linked with X.                                   */

void x_setup(void)
{
    DWORD lin_addr;
    /* Memorize some basic real mode stuff */
    real_code = get_CS();
    real_data = get_DS();
    real_pic1_mask = inp(0x21);
    real_pic2_mask = inp(0xA1);
    /* Insert fundamental descriptors */
    ins_descr(NULL_SEL,0,0,0,0);
    ins_descr(X_VIDEO_SEL,0xB8000,0xFFFF,DATA_ACCESS,GRAN_16);
    ins_descr(X_FLATDATA_SEL,0,0xFFFFF,DATA_ACCESS,GRAN_32B);
    ins_descr(X_FLATCODE_SEL,0,0xFFFFF,CODE_ACCESS,GRAN_32B);
    /* Real address are translated as flat address      */
    /* by following formula: SEG << 4 + OFFSET          */
    lin_addr = ((DWORD)(real_data) << 4);
    ins_descr(X_DATA_SEL,lin_addr,0xFFFF,DATA_ACCESS,GRAN_16);
    lin_addr = ((DWORD)(real_code) << 4);
    ins_descr(X_CODE16_SEL,lin_addr,0xFFFF,CODE_ACCESS,GRAN_16);
    /* The X_CODE32_SELector starts at x_call32 function boundary */
    lin_addr = LINEARIZE(_x_call32_ptr);
    ins_descr(X_CODE32_SEL,lin_addr,0xFFFFF,CODE_ACCESS,GRAN_32);
    
    /* This is the X_info structure access point */
    lin_addr = LINEARIZE(&X_info);
    ins_descr(X_INFO_SEL,lin_addr,sizeof(X_info),DATA_ACCESS,GRAN_16);
    /* This is the gate used to come back to RM */
    lin_addr = LINEARIZE(_x_ret16_ptr)-LINEARIZE(_x_call32_ptr);
    ins_gate(X_RM_BACK_GATE,GDT,X_CODE32_SEL,lin_addr,CALL_GATE386,0);
    /* This is the gate to invoke BIOS */
    lin_addr = LINEARIZE(_x_callBIOS_ptr)-LINEARIZE(_x_call32_ptr);
    ins_gate(X_CALLBIOS_GATE,GDT,X_CODE32_SEL,lin_addr,CALL_GATE386,0);
    /* Pass the GDT & IDT flat address */
    X_info.GDT_base = LINEARIZE(GDT);
    X_info.IDT_base = LINEARIZE(IDT);
}

/* These functions are used to check if the system can run X       */
/* We just check that no EMM386 is installed, CPU is 386 or higher */
/* and it is working in real mode!                                 */

#define CPU_IN_PM       1
#define CPU_NO_386      2
#define CPU_OK          0

unsigned unsupported_32bit(void);
#pragma aux unsupported_32bit = \
	".286P" \
	"pushf" \
	"pop bx" \
	"and bx,0fffh" \
	"push bx" \ 
	"popf" \
	"pushf" \
	"pop ax" \
	"and ax,0f000h" \
	"cmp ax,0f000h" \
	"jz __bad_cpu" \
	"or bx,0f000h" \
	"push bx" \
	"popf" \
	"pushf" \
	"pop ax" \
	"and ax,0f000h" \
	"jz __bad_cpu" \
	"smsw ax" \
	"test ax,1" \
	"jnz __bad_mode" \
	"mov ax,0" \
	"jmp __end" \
	"__bad_mode: mov ax,1" \
	"jmp __end" \
	"__bad_cpu: mov ax,2" \
	"__end: nop" \
	modify [ax bx] \
	value [ax];

const char EMM1[] = "EMMXXXX0";

unsigned EMM_present(void);
#pragma aux EMM_present = \
	"push ds" \
	"mov ax,seg EMM1" \
	"mov ds,ax" \
	"mov dx,offset EMM1" \
	"mov ax,3d00h" \
	"int 21h" \
	"jc __no" \
	"push ax" \
	"mov bx,ax" \
	"mov ax,04400h"  \
	"int 21h" \
	"jc __no_close" \
	"test dx,080h" \
	"jz __no_close" \
	"pop bx" \
	"mov ax,03e00h" \
	"int 21h" \
	"mov ax,1" \
	"jmp __end" \
	"__no_close: pop bx" \
	"mov ax,03e00h" \
	"int 21h" \
	"__no: mov ax,0" \
	"__end: pop ds" \
	modify [ax bx dx] \
	value [ax];

/* Why there is the necessity for a mouse reset ?                        */
/* Well, reprogramming the PIC in PM or installing our own mouse handler */
/* will stuck the existing mouse driver. Anyway calling the function 0H  */
/* of the mouse driver will reset & reactivate it!                       */

void mouse_reset(void);
#pragma aux mouse_reset = \
	"mov ax,0" \
	"int 033H";

/* Similar questions applies from keyboard management! If a key is      */
/* pressed while the application exits, it could happen that a key-code */
/* is lost; this messes up the BIOS handler and keyboard hangs up!      */

void keyb_reset(void)
{
    int dato;
    _disable();
    /* Send the reset command */
    outp(0x60,0xff);
    /* Wait for ack 1 */
    while (!(inp(0x64) & 0x01));
    dato = inp(0x60);
    /* Wait for ack 2 */
    while (!(inp(0x64) & 0x01));
    dato = inp(0x60);
    _enable();
}

void x_checksystem(void)
{
 switch (unsupported_32bit()) {
     case CPU_IN_PM  : printf("Error : Cpu is already working in protected mode");
		       exit(-1);
     case CPU_NO_386 : printf("Error Cpu is 286 or lower model");
		       exit(-1);
     case CPU_OK     : break;
 }
 if (EMM_present()) {
     printf("Error : Found EMM386 driver! Please disable it before running X\n");
     exit(-1);
 }
}

/* This functions is used to check if the application binary file */
/* exists and if it is a valid executable                         */

#define NOEXIST_FILE    0x00
#define ELF_FILE        0x01
#define COFF_FILE       0x02
#define DOS16_FILE      0x03
#define UNKNOWN_FILE    0xFF

/* The file type checking is based on the file initial signature.  */
/* DOS16,ELF & COFF (and hopefully any other executable format)    */
/* are marked at beginning with special byte sequence which allow  */
/* easy identification                                             */
/* If you plan to add another loader you have to follow the line   */
/* given into X_COFF.C & X_ELF.C to implement the loader           */
/* Moreover you have to add here the magic string which identifies */
/* the executable format and add another #define for it            */

unsigned x_checkfile(char *filename)
{
    char signature[16];
    int fh = open(filename,O_RDONLY | O_BINARY);
    int ret;
    if (fh == -1) return(NOEXIST_FILE);
    read(fh,signature,16);
    if ((ret = memcmp(signature,"MZ",2)) == 0) ret = DOS16_FILE;
    else if ((ret = memcmp(&signature[1],"ELF",3)) == 0) ret = ELF_FILE;
    else if ((ret = memcmp(signature,"\x4c\x01",2)) == 0) ret = COFF_FILE;
    else ret = UNKNOWN_FILE;
    close(fh);
    return(ret);
}

/* This is the BIOS call stub; it retrieves     */
/* the information from the x_info structure    */

void x_bios32(void)
{
    union REGS in,out;
    struct SREGS s;
    #ifdef __XBIOS_DEBUG__
	 char __far *nome;
    #endif
	memcpy(&in,&X_info.ir,sizeof(union REGS));
	memcpy(&s,&X_info.sr,sizeof(struct SREGS));
    #ifdef __XBIOS_DEBUG__
	if (in.h.ah == 0x3d) {
	    nome = MK_FP(s.ds,in.x.dx);
	    cprintf("Name : %Fs (%Fp)\n\r",nome,nome);
	    cprintf("AX : %x BX : %x CX : %x DX : %x\n\r",X_info.ir.x.ax,X_info.ir.x.bx,X_info.ir.x.cx,X_info.ir.x.dx);
	    cprintf("SI : %x DI : %x CFLAG : %x\n\r",X_info.ir.x.si,X_info.ir.x.di,X_info.ir.x.cflag);
	    cprintf("CS : %x DS : %x ES : %x SS : %x\n\r",X_info.sr.cs,X_info.sr.ds,X_info.sr.es,X_info.sr.ss);
	    cprintf("AX : %x BX : %x CX : %x DX : %x\n\r",in.x.ax,in.x.bx,in.x.cx,in.x.dx);
	    cprintf("SI : %x DI : %x CFLAG : %x\n\r",in.x.si,in.x.di,in.x.cflag);
	    cprintf("CS : %x DS : %x ES : %x SS : %x\n\r",s.cs,s.ds,s.es,s.ss);
	}
    #endif
    int86x(X_info.irqno,&in,&out,&s);
    memcpy(&X_info.or,&out,sizeof(union REGS));
    memcpy(&s,&X_info.sr,sizeof(struct SREGS));
    #ifdef __XBIOS_DEBUG__
	if (in.h.ah == 0x3d) {
	    cprintf("AX : %x BX : %x CX : %x DX : %x\n\r",X_info.or.x.ax,X_info.or.x.bx,X_info.or.x.cx,X_info.or.x.dx);
	    cprintf("SI : %x DI : %x CFLAG : %x\n\r",X_info.or.x.si,X_info.or.x.di,X_info.or.x.cflag);
	    cprintf("CS : %x DS : %x ES : %x SS : %x\n\r",X_info.sr.cs,X_info.sr.ds,X_info.sr.es,X_info.sr.ss);
	}
    #endif
}

/* The main() checks if the system can support X initializes system     */
/* tables & memory managers and finally checks file & loads the         */
/* application passing the control to it; the application can resume to */
/* X (and real mode) calling the appropriate function through the       */
/* X_RM_BACK_GATE. The HARTIK run-time library provides all the stuff   */
/* needed to do everything easy.                                        */

/* You must should use pseudo-flat addressing; this means that the      */
/* program code & data are not relocated, so their virtual address      */
/* starts from 0; when you load them into extended memory, a segment is */
/* allocated; the segment base is set to the start of the memory block  */
/* and the limit is set to maximum! So the application sees a flat      */
/* space starting from where the application itself has been loaded!    */
/* In order to access low-memory ( < 1 MB area), we need to use far     */
/* pointers whose selector is set to X_FLATCODE_SEL or X_FLATDATA_SEL   */
/* Access to video memory is provided through X_VIDEO_SEL!              */

void main(int argc,char *argv[])
{
    char *filename = NULL;
    char *techstr[] = {"","AT Style","Ps/2 Style"};
    register long i;

    x_checksystem();
    x_setup();
    x_MM_init();
    PM_esp = FP_OFF(&PM_stack[PM_STACK_SIZE]);
    X_info.ver = __X_VERSION__;
    /* Fill the argument passing fields of the info structure */
    /* The memory fields will be filled later after the PM    */
    /* program has been loaded into memory                    */
    /* See X_ELF.C for details                                */
    if (argc < 11) X_info.argc = argc-1;
    else X_info.argc = MAX_ARGS;
    for (i = 1; i < argc && i < MAX_ARGS+1; i++) 
	strncpy(&(X_info.argv[i-1][0]),argv[i],80);
    if (argc > MAX_ARGS)
       printf("Warning\nTrying to pass more than MAX_ARGS(%d) paramaters\n",MAX_ARGS);
    
    if (argc < 2) {
	printf("X! The HARTIK+ Custom extender\n");
	printf("X Version                 : %d.%d\n",(X_info.ver & 0xFF00) >> 8,X_info.ver & 0xFF);
	printf("XMS supported             : %s\n",MM.mode == MM_XMS ? "YES":"NO");
	if (MM.mode == MM_RAW) {                    
	    printf("A20 switch technique      : %s\n",techstr[MM.technique]);
	}
	printf("Conventional memory found : %lu KB\n",MM.dos_size/1024);
	if (MM.mode >= MM_RAW)
	   printf("Extended memory found     : %lu KB\n",MM.extended_size/1024);
	else printf("Warning! Extended memory is not available\n");
	printf("X_info.GDT Base           : %lx\n",X_info.GDT_base);
	exit(0);
    }
    else filename = argv[1];
    switch (x_checkfile(filename)) {
	case ELF_FILE     : ELF_spawn(filename);
			    //mouse_reset();
			    //keyb_reset();
			    break;
	case COFF_FILE    : COFF_spawn(filename);
			    //mouse_reset();
			    //keyb_reset();
			    break;
	case NOEXIST_FILE : printf("Error : File %s does not exist\n",filename);
			    exit(-1);
			    break;
	case UNKNOWN_FILE :
	default           : printf("Error : File %s has an unknown executable format\n",filename);
			    exit(-1);
			    break;
    }    
 exit(1);
}

